from buildings_bench.evaluation.metrics import Metric, MetricType
from buildings_bench.evaluation.metrics import absolute_error, squared_error, bias_error
from buildings_bench.evaluation.scoring_rules import ScoringRule
from buildings_bench.evaluation.scoring_rules import RankedProbabilityScore
from buildings_bench.evaluation.scoring_rules import ContinuousRankedProbabilityScore
from typing import List


metrics_registry = [
    'rmse',
    'mae',
    'nrmse',
    'nmae',
    'mbe',
    'nmbe',
    'cvrmse'
]

scoring_rule_registry = [
    'rps',
    'crps'
]


def metrics_factory(name: str,
                    types: List[MetricType] = [MetricType.SCALAR]) -> List[Metric]:
    """
    Create a metric from a name.
    By default, will return a scalar metric.

    Args:
        name (str): The name of the metric.
        types (List[MetricTypes]): The types of the metric.         
    Returns:
        metrics_list (List[Metric]): A list of metrics. 
    """
    assert name.lower() in metrics_registry, f'Invalid metric name: {name}'

    if name.lower() == 'rmse':
        return [ Metric(f'{name.lower()}-{type}', type, squared_error, sqrt=True) for type in types ]
    elif name.lower() == 'mae':
        return [ Metric(f'{name.lower()}-{type}', type, absolute_error) for type in types ]
    elif name.lower() == 'nrmse':
        return [ Metric(f'{name.lower()}-{type}', type, squared_error, normalize=True, sqrt=True) for type in types ]
    elif name.lower() == 'nmae':
        return [ Metric(f'{name.lower()}-{type}', type, absolute_error, normalize=True) for type in types ]
    elif name.lower() == 'mbe':
        return [ Metric(f'{name.lower()}-{type}', type, bias_error) for type in types ]
    elif name.lower() == 'nmbe':
        return [ Metric(f'{name.lower()}-{type}', type, bias_error, normalize=True) for type in types ]
    elif name.lower() == 'cvrmse':
        return [ Metric(f'{name.lower()}-{type}', type, squared_error, normalize=True, sqrt=True) for type in types ]


def scoring_rule_factory(name: str) -> ScoringRule:
    """Create a scoring rule from a name.

    Args:
        name (str): The name of the scoring rule.
    Returns:
        sr (ScoringRule): A scoring rule.
    """
    assert name.lower() in scoring_rule_registry, f'Invalid scoring rule name: {name}'
    
    if name.lower() == 'crps':
        return ContinuousRankedProbabilityScore()
    elif name.lower() == 'rps':
        return RankedProbabilityScore()


def all_metrics_list() -> List[Metric]:
    """Returns all registered metrics.
    
    Returns:
        metrics_list (List[Metric]): A list of metrics.
    """
    metrics_list = []
    for metric in metrics_registry:
        metrics_list += metrics_factory(metric, types=[MetricType.SCALAR, MetricType.HOUR_OF_DAY])
    return metrics_list



import torch
from buildings_bench.evaluation import metrics_factory
from buildings_bench.evaluation.metrics import Metric, MetricType
from buildings_bench.evaluation.scoring_rules import ScoringRule
from typing import List, Optional
import pandas as pd
from copy import deepcopy


class BuildingTypes:
    """Enum for supported types of buildings.
    
    Attributes:
        RESIDENTIAL (str): Residential building type.
        COMMERCIAL (str): Commercial building type.
        RESIDENTIAL_INT (int): Integer representation of residential building type (0).
        COMMERCIAL_INT (int): Integer representation of commercial building type (1).
    """
    RESIDENTIAL = 'residential'
    COMMERCIAL = 'commercial'
    RESIDENTIAL_INT = 0
    COMMERCIAL_INT = 1


class MetricsManager:
    """A class that keeps track of all metrics (and a scoring rule)for one or more buildings.
    
    Metrics are computed for each building type (residential and commercial).

    Example:

    ```python
    from buildings_bench.evaluation.managers import MetricsManager
    from buildings_bench.evaluation import metrics_factory
    from buildings_bench import BuildingTypes
    import torch


    metrics_manager = MetricsManager(metrics=metrics_factory('cvrmse'))

    metrics_manager(
        y_true=torch.FloatTensor([1, 2, 3]).view(1,3,1),
        y_pred=torch.FloatTensor([1, 2, 3]).view(1,3,1),
        building_type = BuildingTypes.RESIDENTIAL_INT
    )

    for metric in metrics_manager.metrics[BuildingTypes.RESIDENTIAL]:
        metric.mean()
        print(metric.value) # prints tensor(0.)

    ```
    """
    def __init__(self, metrics: List[Metric] = None, scoring_rule: ScoringRule = None):
        """Initializes the MetricsManager.
        
        Args:
            metrics (List[Metric]): A list of metrics to compute for each
                building type.
            scoring_rule (ScoringRule): A scoring rule to compute for each
                building type.
        """
        self.metrics = {}
        if not metrics is None:
            self.metrics[BuildingTypes.RESIDENTIAL] = metrics
            self.metrics[BuildingTypes.COMMERCIAL] = deepcopy(metrics)
        self.scoring_rules = {}
        if not scoring_rule is None:
            self.scoring_rules[BuildingTypes.RESIDENTIAL] = scoring_rule
            self.scoring_rules[BuildingTypes.COMMERCIAL] = deepcopy(scoring_rule)
        self.accumulated_unnormalized_loss = 0
        self.total_samples = 0

    def _compute_all(self, y_true: torch.Tensor,
                     y_pred: torch.Tensor,
                     building_types_mask: torch.Tensor, **kwargs) -> None:
        """Computes all metrics and scoring rules for the given batch.
        
        Args:
            y_true (torch.Tensor): A tensor of shape [batch, pred_len, 1] with
                the true values.
            y_pred (torch.Tensor): A tensor of shape [batch, pred_len, 1] with
                the predicted values.
            building_types_mask (torch.Tensor): A tensor of shape [batch] with  
                the building types of each sample.

        """
        if len(self.metrics) > 0:
            for building_type in [BuildingTypes.RESIDENTIAL, BuildingTypes.COMMERCIAL]:
                
                if building_type == BuildingTypes.RESIDENTIAL:
                    
                    predictions = y_pred[~building_types_mask]
                    if predictions.shape[0] == 0:
                        continue
                    targets = y_true[~building_types_mask]
                elif building_type == BuildingTypes.COMMERCIAL:
                    predictions = y_pred[building_types_mask]
                    if predictions.shape[0] == 0:
                        continue
                    targets = y_true[building_types_mask]

                for metric in self.metrics[building_type]:
                    metric(targets, predictions)
        
        if len(self.scoring_rules) > 0:
            self._compute_scoring_rule(y_true,
                                      kwargs['y_categories'],
                                      kwargs['y_distribution_params'],
                                      kwargs['centroids'],
                                      building_types_mask)


    def _compute_scoring_rule(self, 
                            true_continuous,
                            true_categories,
                            y_distribution_params,
                            centroids,
                            building_types_mask) -> None:
        """Compute the scoring rule.
        
        Args:
            true_continuous (torch.Tensor): The true continuous load values.
                [bsz_sz, pred_len, 1]
            true_categories (torch.Tensor): The true quantized load values.
            y_distribution_params are [bsz_sz, pred_len, vocab_size] if logits,
             or are [bsz_sz, pred_len, 2] if Gaussian
            centroids (torch.Tensor): The bin values for the quantized distribution.
            building_types_mask (torch.Tensor): 
        """
        if y_distribution_params is None:
            raise ValueError('y_distribution_params must be provided to compute scoring rule.')
        
        for building_type in [BuildingTypes.RESIDENTIAL, BuildingTypes.COMMERCIAL]:
            if building_type == BuildingTypes.RESIDENTIAL:
                true_continuous_by_type = true_continuous[~building_types_mask]
                if true_continuous_by_type.shape[0] == 0:
                    continue
                self.scoring_rules[building_type](
                    true_continuous_by_type,
                    true_categories[~building_types_mask],
                    y_distribution_params[~building_types_mask],
                    centroids)
            elif building_type == BuildingTypes.COMMERCIAL:
                true_continuous_by_type = true_continuous[building_types_mask]
                if true_continuous_by_type.shape[0] == 0:
                    continue
                self.scoring_rules[building_type](
                    true_continuous_by_type,
                    true_categories[building_types_mask],
                    y_distribution_params[building_types_mask],
                    centroids)

    def _update_loss(self, loss, sample_size):
        """Updates the accumulated loss and total samples."""
        self.accumulated_unnormalized_loss += (loss * sample_size)
        self.total_samples += sample_size

    def get_ppl(self):
        """Returns the perplexity of the accumulated loss."""
        return torch.exp(self.accumulated_unnormalized_loss) / self.total_samples
    
    def summary(self, with_loss=False, with_ppl=False):
        """Return a summary of the metrics for the dataset.
        
        A summary maps keys to objects of type Metric or ScoringRule.
        """
        summary = {}
        for building_type in [BuildingTypes.RESIDENTIAL, BuildingTypes.COMMERCIAL]:
            summary[building_type] = {}
            if len(self.metrics) > 0:
                for metric in self.metrics[building_type]:
                    if not metric.UNUSED_FLAG:
                        metric.mean()
                        summary[building_type][metric.name] = metric
            if len(self.scoring_rules) > 0: 
                if not self.scoring_rules[building_type].value is None: 
                    self.scoring_rules[building_type].mean()
                    summary[building_type][self.scoring_rules[building_type].name] = \
                        self.scoring_rules[building_type]
                
        if with_ppl and self.total_samples > 0:
            summary['ppl'] = self.get_ppl()
        if with_loss and self.total_samples > 0:
            summary['loss'] = self.accumulated_unnormalized_loss / self.total_samples
        return summary

    def reset(self, loss: bool = True) -> None:
        """Reset the metrics."""
        for building_type in [BuildingTypes.RESIDENTIAL, BuildingTypes.COMMERCIAL]:
            for metric in self.metrics[building_type]:
                metric.reset()
            if len(self.scoring_rules) > 0:
                self.scoring_rules[building_type].reset()
        if loss:
            self.accumulated_unnormalized_loss = 0
            self.total_samples = 0


    def __call__(self, 
                 y_true: torch.Tensor,
                 y_pred: torch.Tensor,
                 building_types_mask: torch.Tensor = None,
                 building_type: int = BuildingTypes.COMMERCIAL_INT,
                 **kwargs):
        """Compute metrics for a batch of predictions.
        
        Args:
            y_true (torch.Tensor): The true (unscaled) load values. (continuous)
                shape is [batch_size, pred_len, 1]
            y_pred (torch.Tensor): The predicted (unscaled) load values. (continuous)
                shape is [batch_size, pred_len, 1]
            building_types_mask (torch.Tensor): 
                A boolean mask indicating the building type of each building.
                True (1) if commercial, False (0). Shape is [batch_size].
            building_type (int): The building type of the batch. Can be provided 
                instead of building_types_mask if all buildings are of the same type.

        Keyword args:
            y_categories (torch.Tensor): The true load values. (quantized)
            y_distribution_params (torch.Tensor): logits, Gaussian params, etc.
            centroids (torch.Tensor): The bin values for the quantized load.
            loss (torch.Tensor): The loss for the batch.
        """
        if building_types_mask is None:
            building_types_mask = (building_type == BuildingTypes.COMMERCIAL_INT) * \
                                   torch.ones(y_true.shape[0], dtype=torch.bool, device=y_true.device)

        self._compute_all(y_true, y_pred, building_types_mask, **kwargs)
        if 'loss' in kwargs:
            batch_size, pred_len, _ = y_true.shape
            self._update_loss(kwargs['loss'], batch_size * pred_len)


class DatasetMetricsManager:
    """
    A class that manages a MetricsManager for each building
    in one or more benchmark datasets. 
    One DatasetMetricsManager can be used to keep track of all metrics
    when evaluating a model on all of the benchmark's datasets.

    This class wil create a Pandas Dataframe summary containing the metrics for each building.

    Default metrics are NRMSE (CVRMSE), NMAE, NMBE.
    """
    default_metrics = metrics_factory('cvrmse',
                      types=[MetricType.SCALAR, MetricType.HOUR_OF_DAY]) \
                      + metrics_factory('nmbe', types=[MetricType.SCALAR, MetricType.HOUR_OF_DAY]) \
                      + metrics_factory('nmae', types=[MetricType.SCALAR, MetricType.HOUR_OF_DAY])

    def __init__(self, 
                 metrics: List[Metric] = default_metrics,
                 scoring_rule: ScoringRule = None):
        """
        Args:
            metrics (List[Metric]): A list of metrics to compute for each
                building type.
            scoring_rule (ScoringRule): A scoring rule to compute for each
                building type.
        """
        self.metrics_list = metrics
        self.scoring_rule = scoring_rule
        self._metrics = {}

    def get_building_from_dataset(self, dataset_name: str, building_id: str) -> Optional[MetricsManager]:
        """If the dataset and building exist, return the MetricsManager for the building.

        Args:
            dataset_name (str): The name of the dataset.
            building_id (str): The unique building identifier.
        
        Returns:
            A MetricsManager if the dataset and building exist, otherwise None.
        """
        # Check if dataset exists
        if dataset_name not in self._metrics:
            return None
        # Check if building exists
        if building_id not in self._metrics[dataset_name]:
            return None
        return self._metrics[dataset_name][building_id]
    
    def add_building_to_dataset_if_missing(self, dataset_name: str, building_id: str) -> None:
        """If the building does not exist, add a new MetricsManager for the building.

        Args:
            dataset_name (str): The name of the dataset.
            building_id (str): The unique building identifier.
        """
        # Check if dataset exists
        if dataset_name not in self._metrics:
            self._metrics[dataset_name] = {}
        # Check if building already exists
        if building_id not in self._metrics[dataset_name]:
            # Use deepcopy to pass new Metric and Scoring Rule objects
            self._metrics[dataset_name][building_id] = \
                MetricsManager(deepcopy(self.metrics_list), deepcopy(self.scoring_rule))


    def summary(self, dataset_name: str = None) -> pd.DataFrame:
        """Return a summary of the metrics for the dataset.
        
        Args:
            dataset_name (str): The name of the dataset to summarize. If None,
                summarize all datasets.
        Returns:
            A Pandas dataframe with the following columns:

                - dataset: The name of the dataset.
                - building_id: The unique ID of the building.
                - building_type: The type of the building.
                - metric: The name of the metric.
                - metric_type: The type of the metric. (scalar or hour_of_day)
                - value: The value of the metric.
        """
        summary = {}
        if dataset_name is None: # summarize all datasets
            for dataset_name in self._metrics.keys():
                summary[dataset_name] = {}
                for building_id in self._metrics[dataset_name].keys():
                    summary[dataset_name][building_id] = \
                        self._metrics[dataset_name][building_id].summary()
        else:
            summary[dataset_name] = {}
            for building_id in self._metrics[dataset_name].keys():
                summary[dataset_name][building_id] = \
                    self._metrics[dataset_name][building_id].summary()
       
        # to Pandas dataframe
        columns = ['dataset', 'building_id', 'building_type', 'metric', 'metric_type', 'value']
        rows = []
        # for each dataset
        for dataset_name in summary.keys():
            # for each building
            for building_id in summary[dataset_name].keys():
                # for the building type 
                for building_type in summary[dataset_name][building_id].keys():
                    # for each metric
                    for metric_name in summary[dataset_name][building_id][building_type].keys():
                        # if scoring rule, skip
                        if self.scoring_rule and metric_name == self.scoring_rule.name:
                            continue
                        # if the metric is a scalar
                        if summary[dataset_name][building_id][building_type][metric_name].type == MetricType.SCALAR:
                            rows.append([dataset_name, building_id, building_type,
                                         metric_name.split('-')[0], MetricType.SCALAR,
                                         summary[dataset_name][building_id][building_type][metric_name].value.item()])
                        # if the metric is a list of scalars
                        elif summary[dataset_name][building_id][building_type][metric_name].type == MetricType.HOUR_OF_DAY:
                            multi_hour_value = summary[dataset_name][building_id][building_type][metric_name].value
                            for hour in range(multi_hour_value.shape[0]):
                                rows.append([dataset_name, building_id,
                                             building_type, metric_name.split('-')[0] + '_' + str(hour), 
                                             MetricType.HOUR_OF_DAY, multi_hour_value[hour].item()])
        
        metric_df = pd.DataFrame(rows, columns=columns)

        if self.scoring_rule:
            columns = ['dataset', 'building_id', 'building_type', 'scoring_rule', 'value']
            rows = []
            for dataset_name in summary.keys():
                for building_id in summary[dataset_name].keys():
                    for building_type in summary[dataset_name][building_id].keys():
                        if self.scoring_rule.name in summary[dataset_name][building_id][building_type]:
                            score = summary[dataset_name][building_id][building_type][self.scoring_rule.name].value
                            for hour in range(score.shape[0]):
                                rows.append([dataset_name, building_id, building_type,
                                            self.scoring_rule.name + '_' + str(hour), score[hour].item()])
            scoring_rule_df = pd.DataFrame(rows, columns=columns)
            return metric_df, scoring_rule_df
        else:
            return metric_df


    def __call__(self, 
                 dataset_name: str,
                 building_id: str,
                 y_true: torch.Tensor, 
                 y_pred: torch.Tensor,
                 building_types_mask: torch.Tensor = None,
                 building_type: int = BuildingTypes.COMMERCIAL_INT,
                 **kwargs) -> None:
        """Compute metrics for a batch of predictions for a single building in a dataset.

        Args:
            dataset_name (str): The name of the dataset.
            building_id (str): The unique building identifier.
            y_true (torch.Tensor): The true (unscaled) load values. (continuous)
                shape is [batch_size, pred_len, 1]
            y_pred (torch.Tensor): The predicted (unscaled) load values. (continuous)
                shape is [batch_size, pred_len, 1]
            building_types_mask (torch.Tensor): 
                A boolean mask indicating the building type of each building.
                True (1) if commercial, False (0). Shape is [batch_size]. Default is None.
            building_type (int): The building type of the batch. Can be provided 
                instead of building_types_mask if all buildings are of the same type.

        Keyword args:
            y_categories (torch.Tensor): The true load values. (quantized)
            y_distribution_params (torch.Tensor): logits, Gaussian params, etc.
            centroids (torch.Tensor): The bin values for the quantized load.
            loss (torch.Tensor): The loss for the batch.        
        """
        self.add_building_to_dataset_if_missing(dataset_name, building_id)
        
        if building_types_mask is None:
            building_types_mask = (building_type == BuildingTypes.COMMERCIAL_INT) * \
                                   torch.ones(y_true.shape[0], dtype=torch.bool, device=y_true.device)
        self._metrics[dataset_name][building_id](y_true, y_pred, building_types_mask, **kwargs)

    



# buildings_bench.models
import torch
from typing import Callable, Tuple, Dict

# Import models here
from buildings_bench.models.dlinear_regression import DLinearRegression
from buildings_bench.models.linear_regression import LinearRegression
from buildings_bench.models.transformers import LoadForecastingTransformer
from buildings_bench.models.persistence import *
from buildings_bench.models.deep_rnn import DeepAutoregressiveRNN


model_registry = {
    'TransformerWithTokenizer-L': LoadForecastingTransformer,
    'TransformerWithTokenizer-M': LoadForecastingTransformer,
    'TransformerWithTokenizer-S': LoadForecastingTransformer,
    'TransformerWithTokenizer-L-ignore-spatial': LoadForecastingTransformer,
    'TransformerWithTokenizer-L-8192': LoadForecastingTransformer,
    'TransformerWithTokenizer-L-474': LoadForecastingTransformer,
    'TransformerWithMSE': LoadForecastingTransformer,
    'TransformerWithGaussian-L': LoadForecastingTransformer,
    'TransformerWithGaussian-M': LoadForecastingTransformer,
    'TransformerWithGaussian-S': LoadForecastingTransformer,
    'AveragePersistence': AveragePersistence,
    'CopyLastDayPersistence': CopyLastDayPersistence,
    'CopyLastWeekPersistence': CopyLastWeekPersistence,
    'LinearRegression': LinearRegression,
    'DLinearRegression': DLinearRegression,
    'DeepAutoregressiveRNN': DeepAutoregressiveRNN,
    # Register your model here
}


def model_factory(model_name: str, model_args: Dict) -> Tuple[torch.nn.Module, Callable, Callable]:
    """Instantiate and returns a model for the benchmark.

    Returns the model itself,
    the loss function to use, and the predict function.

    The predict function should return a tuple of two tensors: 
    (point predictions, prediction distribution parameters) where
    the distribution parameters may be, e.g., logits, or mean and variance.

    Args:
        model_name (str): Name of the model.
        model_args (Dict): The keyword arguments for the model.
    Returns:
        model (torch.nn.Module): the instantiated model  
        loss (Callable): loss function
        predict (Callable): predict function
    """
    assert model_name in model_registry.keys(), \
        f"Model {model_name} not in registry: {model_registry.keys()}"
    
    model = model_registry[model_name](**model_args)
    loss = model.loss
    predict = model.predict
    return model, loss, predict




import unittest
from buildings_bench import tokenizer
import os
from pathlib import Path 
import numpy as np


class TestTokenizer(unittest.TestCase):
    """ Test the KMeans tokenizer"""

    def test_load(self):
        """ Test loading the tokenizer """

        transform_path = Path(os.environ.get('BUILDINGS_BENCH', '')) \
                              / 'metadata' / 'transforms'

        load_quantizer = tokenizer.LoadQuantizer(with_merge=True, num_centroids=3747, device='cpu')
        load_quantizer.load(transform_path)
        self.assertEqual(load_quantizer.get_vocab_size(), 3747)


    def test_transform_cpu(self):
        """ Test the transform method on CPU."""

        transform_path = Path(os.environ.get('BUILDINGS_BENCH', '')) \
                              / 'metadata' / 'transforms'

        load_quantizer = tokenizer.LoadQuantizer(with_merge=True, num_centroids=3747, device='cpu')
        load_quantizer.load(transform_path)

        x = np.array([[100.234], [0.234], [55.523]])
        y = load_quantizer.transform(x)
        z = load_quantizer.undo_transform(y)

        #print(x,y,z)

        self.assertTrue(np.allclose(x, z, atol=1))



import pandas as pd
from pathlib import Path 
from rliable import library as rly
import numpy as np
from buildings_bench import BuildingTypes


def return_aggregate_median(model_list, 
                            results_dir,
                            experiment='zero_shot',
                            metrics=['cvrmse'], 
                            exclude_simulated = True,
                            only_simulated = False,
                            oov_list = [],
                            reps=50000):
    """Compute the aggregate median for a list of models and metrics over all buildings.
    Also returns the stratified 95% boostrap CIs for the aggregate median.

    Args:
        model_list (list): List of models to compute aggregate median for.
        results_dir (str): Path to directory containing results.
        experiment (str, optional): Experiment type. Defaults to 'zero_shot'.
            Options: 'zero_shot', 'transfer_learning'.
        metrics (list, optional): List of metrics to compute aggregate median for. Defaults to ['cvrmse'].
        exclude_simulated (bool, optional): Whether to exclude simulated data. Defaults to True.
        only_simulated (bool, optional): Whether to only include simulated data. Defaults to False.
        oov_list (list, optional): List of OOV buildings to exclude. Defaults to [].
        reps (int, optional): Number of bootstrap replicates to use. Defaults to 50000.

    Returns:
        result_dict (Dict): Dictionary containing aggregate median and CIs for each metric and building type.
    """

    result_dict = {}        
    aggregate_func = lambda x : np.array([
        np.median(x.reshape(-1))])
    for building_type in [BuildingTypes.RESIDENTIAL, BuildingTypes.COMMERCIAL]:
        result_dict[building_type] = {}
        for metric in metrics:
            result_dict[building_type][metric] = {}

            if experiment == 'zero_shot' and (metric == 'rps' or metric == 'crps'):
                prefix = 'scoring_rule'
            elif experiment == 'transfer_learning' and (metric == 'rps' or metric == 'crps'):
                prefix = 'TL_scoring_rule'
            elif experiment == 'zero_shot':
                prefix = 'metrics'
            elif experiment == 'transfer_learning':
                prefix = 'TL_metrics'
    
            for model in model_list:
                df = pd.read_csv(Path(results_dir) / f'{prefix}_{model}.csv')

                if len(oov_list) > 0:
                    # Remove OOV buildings
                    df = df[~df['building_id'].str.contains('|'.join(oov_list))]
                
                if exclude_simulated:
                    # Exclude synthetic data
                    df = df[~( (df['dataset'] == 'buildings-900k-test') | (df['dataset'] == 'buildings-1m-test') )]
                elif only_simulated:
                    df = df[ (df['dataset'] == 'buildings-900k-test') | (df['dataset'] == 'buildings-1m-test') ]
                
                # if any df values are inf or nan
                if df.isnull().values.any() or np.isinf(df.value).values.any():
                    print(f'Warning: {model} has inf/nan values')
                # REmove inf/nan values
                df = df.replace(np.inf, np.nan)
                df = df.dropna() 

                if metric != 'rps' and metric != 'crps':    
                    result_dict[building_type][metric][model] = \
                        df[(df['metric'] == metric) & (df['building_type'] == building_type)]['value'].values.reshape(-1,1)
                else:
                    result_dict[building_type][metric][model] = \
                        df[df['building_type'] == building_type]['value'].values.reshape(-1,1)

            aggregate_scores, aggregate_score_cis = rly.get_interval_estimates(
                result_dict[building_type][metric], aggregate_func, reps=reps)
            result_dict[building_type][metric] = (aggregate_scores, aggregate_score_cis)
    return result_dict


def pretty_print_aggregates(results_dict) -> None:
    """Pretty print the aggregate results. 
    
    `model_name building_type metric_name: value (CI)`

    Args:
        result_dict (Dict): Dictionary of aggregate metrics for residential and commercial buildings.
    """
    print('model name\t building type\t metric name\t value (95% C.I.)')
    print('==========================================================')

    for building_type, v in results_dict.items():
        for metric_name, vv in v.items():
            agg_scores, agg_cis = vv
            for model_name, metric_value in agg_scores.items():
                cis = agg_cis[model_name]
                if metric_name in ['nrmse', 'nmae', 'nmbe', 'cvrmse']:
                    metric_value = metric_value[0]*100
                else:
                    metric_value = metric_value[0]
                print(f'{model_name} {building_type} {metric_name}: {metric_value:.3f} ({cis[0][0]:.3f},{cis[1][0]:.3f})')


if __name__ == '__main__':

    import os

    oov = []
    with open(Path(os.environ.get('BUILDINGS_BENCH', '')) / 'metadata' / 'oov.txt', 'r') as f:
        for l in f:
            oov += [l.strip().split(' ')[1]]

    results_dir = '/data/local/projects/foundation/NREL/BuildingsBench/results/remove_outliers'

    pretty_print_aggregates(return_aggregate_median(['AveragePersistence', 'TransformerWithGaussian-L'],
                             results_dir,
                             experiment='zero_shot',
                             oov_list = oov))


import torch
from buildings_bench.models.base_model import BaseModel


class AveragePersistence(BaseModel):
    """Predict each hour as the average over each previous day.
    """
    def __init__(self, context_len=168, pred_len=24, continuous_loads=True):
        super().__init__(context_len, pred_len, continuous_loads)

    def forward(self, x):
        # [bsz x seqlen x 1]
        src_series = x['load'][:,:self.context_len,:]

        preds = []
        stds = []
        for i in range(self.pred_len):
            preds += [ torch.mean(src_series[:,i::24,:], dim=1) ]
            stds += [ torch.clamp(
                torch.std(src_series[:,i::24,:], dim=1),
                min=1e-3)]
        return torch.cat([torch.stack(preds, 1), torch.stack(stds, 1)],2)
    
    def loss(self, x, y):
        return x, y
   
    def predict(self, x):
        out = self.forward(x)
        return out[:,:,0].unsqueeze(-1), out

    def unfreeze_and_get_parameters_for_finetuning(self):
        return None
    
    def load_from_checkpoint(self, checkpoint_path):
        return None
    

class CopyLastDayPersistence(BaseModel):
    """Predict each hour as the same hour from the previous day.
    """
    def __init__(self, context_len=168, pred_len=24, continuous_loads=True):
        super().__init__(context_len, pred_len, continuous_loads)
        assert self.context_len >= 24
        assert self.pred_len >= 24

    def forward(self, x):
        # [bsz x seqlen x 1]
        src_series = x['load'][:,:self.context_len,:]
        return src_series[:, self.context_len-24:]
    
    def loss(self, x, y):
        return x, y
    
    def predict(self, x):
        return self.forward(x), None

    def unfreeze_and_get_parameters_for_finetuning(self):
        return None
    
    def load_from_checkpoint(self, checkpoint_path):
        return None


class CopyLastWeekPersistence(BaseModel):
    """Predict each hour as the same hour from the previous week.
    """
    def __init__(self, context_len=168, pred_len=24, continuous_loads=True):
        super().__init__(context_len, pred_len, continuous_loads)
        assert self.context_len >= 168
        assert self.pred_len >= 24

    def forward(self, x):
        # [bsz x seqlen x 1]
        src_series = x['load'][:,:self.context_len,:]
        return src_series[:, self.context_len-168 : self.context_len - 168 +24]

    def loss(self, x, y):
        return x, y
    
    def predict(self, x):
        return self.forward(x), None
    
    def unfreeze_and_get_parameters_for_finetuning(self):
        return None
    
    def load_from_checkpoint(self, checkpoint_path):
        return None



from buildings_bench.evaluation.managers import BuildingTypes
from buildings_bench.data import load_torch_dataset, load_pretraining, load_pandas_dataset
from buildings_bench.data import benchmark_registry

__version__ = "0.1.1"


import numpy as np
import random 
import torch 
import os 
import datetime 


def set_seed(seed: int = 42) -> None:
    """Set random seed for reproducibility.
    """
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    #print(f"Random seed set as {seed}")


def save_model_checkpoint(model, optimizer, scheduler, step, path):
    """Save model checkpoint.
    """
    checkpoint = {
        'model': model.state_dict(),
        'optimizer': optimizer.state_dict(),
        'scheduler': scheduler.state_dict(),
        'step': step
    }
    torch.save(checkpoint, path)
    #print(f"Saved model checkpoint to {path}...")


def load_model_checkpoint(path, model, optimizer, scheduler, local_rank):
    """Load model checkpoint.
    """
    checkpoint = torch.load(path, map_location=f'cuda:{local_rank}')
    model.load_state_dict(checkpoint['model'])
    optimizer.load_state_dict(checkpoint['optimizer'])
    scheduler.load_state_dict(checkpoint['scheduler'])
    step = checkpoint['step']
    #print(f"Loaded model checkpoint from {path}")
    return model, optimizer, scheduler, step


def worker_init_fn_eulp(worker_id):
    """Set random seed for each worker and init file pointer
    for Buildings-900K dataset workers.

    Args:
        worker_id (int): worker id
    """
    np.random.seed(np.random.get_state()[1][0] + worker_id)
    worker_info = torch.utils.data.get_worker_info()
    worker_info.dataset.init_fp()
   

def time_features_to_datetime(time_features: np.ndarray,
                              year: int) -> np.array:
    """
    Convert time features to datetime objects.

    Args:
        time_features (np.ndarray): Array of time features.
            [:,0] is day of year,
            [:,1] is day of week,
            [:,2] is hour of day.
        year (int): Year to use for datetime objects.

    Returns:
        np.array: Array of datetime objects.
    """
    day_of_year = time_features[:,0]
    hour_of_day = time_features[:,2]
    return np.array([datetime.datetime(year, 1, 1, 0, 0, 0) +   # January 1st
                    datetime.timedelta(days=int(doy-1), hours=int(hod), minutes=0, seconds=0)
                    for doy, hod in zip(day_of_year, hour_of_day)])   



import numpy as np
from pathlib import Path
import pandas as pd
import torch
import pickle as pkl
import os
import sklearn.preprocessing as preprocessing
from typing import Union

       
class BoxCoxTransform:
    """A class that computes and applies the Box-Cox transform to data.
    """
    def __init__(self, max_datapoints=1000000):
        """
        Args:
            max_datapoints (int): If the number of datapoints is greater than this, subsample.
        """
        self.boxcox = None
        self.max_datapoints = max_datapoints

    def train(self, data: np.array) -> None:
        """Train the Box-Cox transform on the data with sklearn.preprocessing.PowerTransformer.
        
        Args:
            data (np.array): of shape (n, 1) or (b,n,1)
        """       
        self.boxcox = preprocessing.PowerTransformer(method='box-cox', standardize=True)
        data = data.flatten().reshape(-1,1)
        if data.shape[0] > self.max_datapoints:
            #print(f'Box-Cox: subsampling {self.max_datapoints} datapoints')
            data = data[np.random.choice(data.shape[0], self.max_datapoints, replace=False)]
        self.boxcox.fit_transform(1e-6 + data)


    def save(self, output_path: Path) -> None:
        """Save the Box-Cox transform"""
        with open(output_path / "boxcox.pkl", 'wb') as f:
            pkl.dump(self.boxcox, f)
    
    def load(self, saved_path: Path) -> None:
        """Load the Box-Cox transform"""
        with open(saved_path / "boxcox.pkl", 'rb') as f:
            self.boxcox = pkl.load(f)

    def transform(self, sample: np.ndarray) -> np.ndarray:
        """Transform a sample via Box-Cox.
        Not ran on the GPU, so input/output are numpy arrays.

        Args:
            sample (np.ndarray): of shape (n, 1) or (b,n,1) 
        
        Returns:
            transformed_sample (np.ndarray): of shape (n, 1) or (b,n,1)
        """
        init_shape = sample.shape
        return self.boxcox.transform(1e-6 + sample.flatten().reshape(-1,1)).reshape(init_shape)


    def undo_transform(self, sample: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]: 
        """Undo the transformation of a sample via Box-Cox
        
        Args:
            sample (np.ndarray) or (torch.LongTensor): of shape (n, 1) or (b,n,1). 
                numpy if device is cpu or torch Tensor if device is cuda.

        Returns:
            unscaled_sample (np.ndarray or torch.Tensor): of shape (n, 1) or (b,n,1).
        """
        is_tensor = isinstance(sample, torch.Tensor)
        # if torch.Tensor, convert to numpy first
        if is_tensor:
            device = sample.device
            sample = sample.cpu().numpy()
        init_shape = sample.shape       
        sample = self.boxcox.inverse_transform(sample.flatten().reshape(-1,1)).reshape(init_shape)
        # convert back to torch
        if is_tensor:
            sample = torch.from_numpy(sample).to(device)
        return sample
    

class StandardScalerTransform:
    """ A class that standardizes data by removing the mean and scaling to unit variance.
    """
    def __init__(self, max_datapoints=1000000, device='cpu'):
        """
        Args:
            max_datapoints (int): If the number of datapoints is greater than this, subsample.
            device (str): 'cpu' or 'cuda'
        """
        self.mean_ = None
        self.std_ = None
        self.max_datapoints = max_datapoints
        self.device=device

    def train(self, data: np.array) -> None:
        """Train the StandardScaler transform on the data.
        
        Args:
            data (np.array): of shape (n, 1) or (b,n,1)
        """       
        data = data.flatten().reshape(-1,1)
        if data.shape[0] > self.max_datapoints:
            #print(f'Subsampling {self.max_datapoints} datapoints to fit the StandardScalerTransform')
            data = data[np.random.choice(data.shape[0], self.max_datapoints, replace=False)]
        self.mean_ = torch.from_numpy(np.array([np.mean(data)])).float().to(self.device)
        self.std_ = torch.from_numpy(np.array([np.std(data)])).float().to(self.device)


    def save(self, output_path: Path) -> None:
        """Save the StandardScaler transform"""
        mean_ = self.mean_.cpu().numpy().reshape(-1)
        std_ = self.std_.cpu().numpy().reshape(-1)
        np.save(output_path / "standard_scaler.npy", np.array([mean_, std_]))
    
    
    def load(self, saved_path: Path) -> None:
        """Load the StandardScaler transform"""
        x = np.load(saved_path / "standard_scaler.npy")
        self.mean_ = torch.from_numpy(np.array([x[0]])).float().to(self.device)
        self.std_ = torch.from_numpy(np.array([x[1]])).float().to(self.device)


    def transform(self, sample: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """Transform a sample via StandardScaler
        
        Args:
            sample (np.ndarray or torch.Tensor): shape (n, 1) or (b,n,1) 
        Returns:
            transformed_samples (torch.Tensor): shape (n, 1) or (b,n,1)
        """
        if isinstance(sample, np.ndarray):
            sample = torch.from_numpy(sample).float().to(self.device)        
        return (sample - self.mean_) / self.std_


    def undo_transform(self, sample: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
        """Undo the transformation of a sample via StandardScaler
        
        Args:
            sample (np.ndarray): of shape (n, 1) or (b,n,1) or torch.Tensor of shape (n, 1) or (b,n,1)

        Returns:
            unscaled_sample (torch.Tensor): of shape (n, 1) or (b,n,1)
        """
        if isinstance(sample, np.ndarray):
            sample = torch.from_numpy(sample).float().to(self.device)
        return self.std_ * sample + self.mean_        


    def undo_transform_std(self, scaled_std: torch.Tensor) -> torch.Tensor:
        """Undo transform for standard deviation.
        
        Args:
            scaled_std (torch.Tensor): of shape (n, 1) or (b,n,1)
        
        Returns:
            unscaled_std (torch.Tensor): of shape (n, 1) or (b,n,1)
        """
        return self.std_ * scaled_std
    
        
class LatLonTransform:
    """Pre-processing lat,lon data with standard normalization by Buildings-900K training set.
    """
    def __init__(self):
        metadata_path = Path(os.environ.get('BUILDINGS_BENCH','')) / 'metadata'
        # Load withheld pumas
        with open(metadata_path / 'withheld_pumas.tsv', 'r') as f:
            # tab separated file
            line = f.readlines()[0]
            self.withheld_pumas = line.split('\t')

        census_regions = ['northeast', 'midwest', 'south', 'west']
        self.puma_to_centroid = {}
        lat_means, lat_stds = 0, 0
        lon_means, lon_stds = 0, 0
        for idx,cr in enumerate(census_regions):
            cr_df = pd.read_csv(metadata_path / f'map_of_pumas_in_census_region_{idx+1}_{cr}.csv', header=0)
            # keys are cr_df['GISJOIN'], values are (cr_df['latitude'], cr_df['longitude'])
            self.puma_to_centroid.update(dict(zip(cr_df['GISJOIN'], zip(cr_df['latitude'], cr_df['longitude']))))
            # Filter out withheld pumas
            non_withheld_df = cr_df[~cr_df['GISJOIN'].isin(self.withheld_pumas)]
            # Compute mean and std 
            lat_means += non_withheld_df['latitude'].mean()
            lat_stds += non_withheld_df['latitude'].std()
            lon_means += non_withheld_df['longitude'].mean()
            lon_stds += non_withheld_df['longitude'].std()

        self.lat_means = lat_means / 4
        self.lon_means = lon_means / 4
        self.lat_stds = lat_stds / 4
        self.lon_stds = lon_stds / 4

        # convert self.puma_to_centroid values to np.ndarray of shape (2,)
        for k,v in self.puma_to_centroid.items():
            # Normalize
            v = (np.array(v, dtype=np.float32) - np.array([self.lat_means, self.lon_means])) / np.array([self.lat_stds, self.lon_stds])
            self.puma_to_centroid[k] = v.astype(np.float32)


    def transform_latlon(self, latlon: np.ndarray) -> np.ndarray:
        """Transform a raw Lat/Lon sample into a normalized Lat/Lon sample

        Args:
            latlon (np.ndarray): of shape (2,).

        Returns:
            transformed_latlon (np.ndarray): of shape (2,).
        """
        return (latlon - np.array([self.lat_means, self.lon_means])) / np.array([self.lat_stds, self.lon_stds])

    def undo_transform(self, normalized_latlon: np.ndarray) -> np.ndarray:
        """Undo the transformation of a sample

        Args:
            normalized_latlon (np.ndarray): of shape (n, 2) or (b,n,2).

        Returns:
            unnormalized_latlon (np.ndarray): of shape (n, 2) or (b,n,2).
        """
        init_shape = normalized_latlon.shape
        normalized_latlon = normalized_latlon.reshape(-1,2)
        lat = normalized_latlon[:,0] * self.lat_stds + self.lat_means
        lon = normalized_latlon[:,1] * self.lon_stds + self.lon_means
        return np.stack([lat, lon], axis=1).reshape(init_shape)

    def transform(self, puma_id: str) -> np.ndarray:
        """Look up a PUMA ID's normalized Lat/Lon centroid.

        This is used in the Buildings-900K Dataset to look up a lat/lon
        for each building's PUMA.

        Args:
            puma_id (str): PUMA ID
        
        Returns:
            centroid (np.ndarray): of shape (1,2)
        """
        return self.puma_to_centroid[puma_id].reshape(1,2)


class TimestampTransform:
    """Extract timestamp features from a Pandas timestamp Series.
    """
    def __init__(self, is_leap_year: bool = False):
        """
        Args:
            is_leap_year (bool): Whether the year of the building data is a leap year or not.
        """
        self.day_year_normalization = 365 if is_leap_year else 364
        self.hour_of_day_normalization = 23
        self.day_of_week_normalization = 6

    def transform(self, timestamp_series: pd.DataFrame) -> np.ndarray:
        """Extract timestamp features from a Pandas timestamp Series.


        - Day of week (0-6)
        - Day of year (0-364)
        - Hour of day (0-23)

        Args:
            timestamp_series (pd.DataFrame): of shape (n,) or (b,n)
        
        Returns:
            time_features (np.ndarray): of shape (n,3) or (b,n,3)
        """
        # If the input is a DatetimeIndex
        if isinstance(timestamp_series, pd.DatetimeIndex):
            timestamp_series = timestamp_series.to_series()
        # Convert to datetime
        timestamp_series = pd.to_datetime(timestamp_series)
        # Extract features
        day_of_week = timestamp_series.dt.dayofweek
        day_of_year = timestamp_series.dt.dayofyear
        hour_of_day = timestamp_series.dt.hour
        time_features = np.stack([day_of_year / self.day_year_normalization,
                         day_of_week / self.day_of_week_normalization,
                         hour_of_day / self.hour_of_day_normalization], axis=1).astype(np.float32)
        return time_features * 2 - 1


    def undo_transform(self, time_features: np.ndarray) -> np.ndarray:
        """Convert normalized time features back to original time features

        Args:
            time_features (np.ndarray): of shape (n, 3) or (b,n,3)
        
        Returns:
            unnormalized_time_features (np.ndarray): of shape (n, 3) or (b,n,3)
        """
        init_shape = time_features.shape
        time_features = time_features.reshape(-1,3)
        time_features = (time_features + 1) * 0.5
        day_of_year = np.round(time_features[:,0] * self.day_year_normalization)
        day_of_week = np.round(time_features[:,1] * self.day_of_week_normalization)
        hour_of_day = np.round(time_features[:,2] * self.hour_of_day_normalization)
        return np.stack([day_of_year, day_of_week, hour_of_day], axis=1).astype(np.int32).reshape(init_shape)



import torch 
import math 
from buildings_bench.evaluation.metrics import MetricType
from buildings_bench.evaluation.metrics import BuildingsBenchMetric


class ScoringRule(BuildingsBenchMetric):
    """An abstract class for all scoring rules.
    """

    def __init__(self, name: str):
        super().__init__(name, MetricType.HOUR_OF_DAY)

    def reset(self):
        self.value = None

    def __call__(self, **kwargs):
        raise NotImplementedError()
    
    def mean(self):
        if self.value is None:
            return
        value = torch.stack(self.value,0)
        self.value = torch.mean(value, 0)    


class RankedProbabilityScore(ScoringRule):
    """A class that calculates the ranked probability score (RPS) metric
    for categorical distributions."""

    def __init__(self):
        super().__init__(name='rps')

    def rps(self, y_true, y_pred_logits, centroids) -> None:
        """A PyTorch method that calculates the ranked probability score metric
           for categorical distributions.

           Since the bin values are centroids of clusters along the real line,
           we have to compute the width of the bins by summing the distance to
           the left and right centroids of the bin (divided by 2), except for
           the first and last bins, where we only need to sum the distance to
           the right centroid of the first bin and the left centroid of the
           last bin, respectively.

        Args:
            y_true (torch.Tensor): of shape [batch_size, seq_len, 1] categorical labels
            y_pred_logits (torch.Tensor): of shape [batch_size, seq_len, vocab_size] logits
            centroids (torch.Tensor): of shape [vocab_size]
        """
        # Convert class labels y_true to one hot vectors [batch_size, seq_len, vocab_size]
        y_true = torch.nn.functional.one_hot(y_true.squeeze(2).long(),
                                             num_classes=centroids.shape[0]).to(y_pred_logits.device)
        # Sort the values, logits, and y_true
        centroids, indices = torch.sort(centroids, dim=-1)
        y_pred_logits = y_pred_logits[:, :, indices]
        y_true = y_true[:, :, indices]

        softmax_preds = torch.softmax(y_pred_logits, dim=-1)
        # Calculate the cumulative distribution function (CDF) of the predictions
        cdf = torch.cumsum(softmax_preds, dim=-1)

        y_true_cdf = torch.cumsum(y_true, dim=-1)
        # Calculate the difference between the CDF and the true values
        square = torch.square(cdf - y_true_cdf)
        
        # Calculate the widths of the bins:
        # we need to calculate
        # half the distance to the right centroid and left centroid.
        centroid_dist = centroids[1:] - centroids[:-1]
        half_dists = centroid_dist / 2
        widths = torch.unsqueeze(half_dists[1:] + half_dists[:-1], dim=0)
        widths = torch.cat([
            centroids[0].view(1, 1) + half_dists[0].view(1, 1),
            widths,
            half_dists[-1].view(1, 1)
        ],dim=1)
        
        # Calculate the RPS    
        rps = torch.mean(torch.sum(square * widths, dim=-1), dim=0)  # [seq_len]
        if self.value is None:
            self.value = [rps]
        else:
            self.value += [rps]

    def __call__(self, true_continuous, y_true, y_pred_logits, centroids):
        self.rps(y_true, y_pred_logits, centroids)


class ContinuousRankedProbabilityScore(ScoringRule):
    """
    A class that calculates the Gaussian continuous ranked probability score (CRPS) metric.
    """
    def __init__(self):
        super().__init__(name = 'crps')

    def crps(self, true_continuous, y_pred_distribution_params) -> None:
        """Computes the Gaussian CRPS.

        Args:
            true_continuous (torch.Tensor): of shape [batch_size, seq_len, 1]
            y_pred_distribution_params (torch.Tensor): of shape [batch_size, seq_len, 2]
        """
        pred_mu = y_pred_distribution_params[:, :, 0].unsqueeze(-1)
        pred_sigma = y_pred_distribution_params[:, :, 1].unsqueeze(-1)
        
        # standardize the true values to N(0,1)
        true_continuous = (true_continuous - pred_mu) / pred_sigma
        # Calculate the cumulative distribution function (CDF) of the predictions
        cdf = 0.5 * (1 + torch.erf(true_continuous / math.sqrt(2)))
        # Calculate the pdf of the predictions
        pdf = torch.exp(-torch.square(true_continuous) / 2) / math.sqrt(2 * math.pi)
        # Calculate pi inv
        pi_inv = 1 / math.sqrt(math.pi)
        # CRPS
        crps = pred_sigma * ( true_continuous * (2 * cdf - 1) + 2 * pdf - pi_inv)
        crps = torch.mean(crps, dim=0).squeeze(1)  # [seq_len]
        if self.value is None:
            self.value = [crps]
        else:
            self.value += [crps]

    def __call__(self, true_continuous, y_true, y_pred_distribution_params, centroids):
        self.crps(true_continuous, y_pred_distribution_params)



import unittest
import torch
import numpy as np

class TestPersistence(unittest.TestCase):
    def test_average_persistence(self):
        from buildings_bench.models.persistence import AveragePersistence

        context_len = 168
        pred_len = 24
        bsz = 10
        seqlen = context_len + pred_len
        x = {'load': torch.from_numpy(np.random.rand(bsz, seqlen, 1).astype(np.float32))}
        ap = AveragePersistence(context_len=context_len, pred_len=pred_len)
        y = ap(x)
        y_mean = y[:, :, 0]
        y_sigma = y[:, :, 1]
        self.assertEqual(y_mean.shape, (bsz, pred_len))
        self.assertEqual(y_sigma.shape, (bsz, pred_len))

    def test_copy_last_persistence(self):
        from buildings_bench.models.persistence import CopyLastDayPersistence

        context_len = 168
        pred_len = 24
        bsz = 10
        seqlen = context_len + pred_len
        load = torch.from_numpy(np.random.rand(bsz, seqlen, 1).astype(np.float32))
        last_day = load[:, -48:-24]
        x = {'load': load}
        ap = CopyLastDayPersistence(context_len=context_len, pred_len=pred_len)
        y = ap(x)
        self.assertEqual(y.shape, (bsz, pred_len, 1))
        self.assertTrue( (last_day == y).all() )

    def test_last_week_persistence(self):
        from buildings_bench.models.persistence import CopyLastWeekPersistence

        context_len = 168
        pred_len = 24
        bsz = 10
        seqlen = context_len + pred_len
        load = torch.from_numpy(np.random.rand(bsz, seqlen, 1).astype(np.float32))
        last_week = load[:,0:24]
        x = {'load': load}
        ap = CopyLastWeekPersistence(context_len=context_len, pred_len=pred_len)
        y = ap(x)
        self.assertEqual(y.shape, (bsz, pred_len, 1))
        self.assertTrue( (last_week == y).all() )

def test():
    suite = unittest.TestLoader().loadTestsFromTestCase(TestPersistence)
    unittest.TextTestRunner(verbosity=2).run(suite)



import torch
import numpy as np
from pathlib import Path
import pyarrow.parquet as pq
import buildings_bench.transforms as transforms
from buildings_bench.transforms import BoxCoxTransform, StandardScalerTransform


class Buildings900K(torch.utils.data.Dataset):
    r"""This is an indexed dataset for the Buildings-900K dataset.
    It uses an index file to quickly load a sub-sequence from a time series in a multi-building
    Parquet file. The index file is a tab separated file with the following columns:

    0. Building-type-and-year (e.g., comstock_tmy3_release_1)
    1. Census region (e.g., by_puma_midwest)
    2. PUMA ID
    3. Building ID
    4. Hour of year pointer (e.g., 0070)

    The sequence pointer is used to extract the slice
    [pointer - context length : pointer + pred length] for a given building ID.
    
    The time series are not stored chronologically and must be sorted by timestamp after loading.

    Each dataloader worker has its own file pointer to the index file. This is to avoid
    weird multiprocessing errors from sharing a file pointer. We 'seek' to the correct
    line in the index file for random access.
    """
    def __init__(self, 
                dataset_path: Path,
                index_file: str,
                context_len: int = 168,
                pred_len: int = 24,
                apply_scaler_transform: str = '',
                scaler_transform_path: Path = None):
        """
        Args:
            dataset_path (Path): Path to the pretraining dataset.
            index_file (str): Name of the index file
            context_len (int, optional): Length of the context. Defaults to 168. 
                The index file has to be generated with the same context length.
            pred_len (int, optional): Length of the prediction horizon. Defaults to 24.
                The index file has to be generated with the same pred length.
            apply_scaler_transform (str, optional): Apply a scaler transform to the load. Defaults to ''.
            scaler_transform_path (Path, optional): Path to the scaler transform. Defaults to None.
        """
        self.dataset_path = dataset_path / 'Buildings-900K' / 'end-use-load-profiles-for-us-building-stock' / '2021'
        self.metadata_path = dataset_path / 'metadata'
        self.context_len = context_len
        self.pred_len = pred_len
        self.building_type_and_year = ['comstock_tmy3_release_1',
                                       'resstock_tmy3_release_1',
                                       'comstock_amy2018_release_1',
                                       'resstock_amy2018_release_1']
        self.census_regions = ['by_puma_midwest', 'by_puma_south', 'by_puma_northeast', 'by_puma_west']
        self.index_file = self.metadata_path / index_file
        self.index_fp = None
        self.__read_index_file(self.index_file)
        self.time_transform = transforms.TimestampTransform()
        self.spatial_transform = transforms.LatLonTransform()
        self.apply_scaler_transform = apply_scaler_transform
        if self.apply_scaler_transform == 'boxcox':
            self.load_transform = BoxCoxTransform()
            self.load_transform.load(scaler_transform_path)
        elif self.apply_scaler_transform == 'standard':
            self.load_transform = StandardScalerTransform()
            self.load_transform.load(scaler_transform_path)


    def init_fp(self):
        """Each worker needs to open its own file pointer to avoid 
        weird multiprocessing errors from sharing a file pointer.

        This is not called in the main process.
        This is called in the DataLoader worker_init_fn.
        The file is opened in binary mode which lets us disable buffering.
        """
        self.index_fp = open(self.index_file, 'rb', buffering=0)
        self.index_fp.seek(0)

    def __read_index_file(self, index_file: Path) -> None:
        """Extract metadata from index file.
        """
        # Fast solution to get the number of time series in index file
        # https://pynative.com/python-count-number-of-lines-in-file/
        def _count_generator(reader):
            b = reader(1024 * 1024)
            while b:
                yield b
                b = reader(1024 * 1024)

        with open(index_file, 'rb') as fp:
            c_generator = _count_generator(fp.raw.read)
            # count each \n
            self.num_time_series = sum(buffer.count(b'\n') for buffer in c_generator)
        
        # Count the number of chars per line
        with open(index_file, 'rb', buffering=0) as fp:
            first_line = fp.readline()
            self.chunk_size = len(first_line)
        
        #print(f'Counted {self.num_time_series} indices in index file.')

    def __del__(self):
        if self.index_fp:
            self.index_fp.close()

    def __len__(self):
        return self.num_time_series

    def __getitem__(self, idx):
        # Open file pointer if not already open
        if not self.index_fp:
           self.index_fp = open(self.index_file, 'rb', buffering=0)
           self.index_fp.seek(0)

        # Get the index of the time series
        self.index_fp.seek(idx * self.chunk_size, 0)
        ts_idx = self.index_fp.read(self.chunk_size).decode('utf-8')

        # Parse the index
        ts_idx = ts_idx.strip('\n').split('\t')

        # strip loading zeros
        seq_ptr = ts_idx[-1]
        seq_ptr = int(seq_ptr.lstrip('0')) if seq_ptr != '0000' else 0
        # Building ID
        bldg_id = ts_idx[3].lstrip('0')

        # Select timestamp and building column
        df = pq.read_table(str(self.dataset_path / self.building_type_and_year[int(ts_idx[0])]
                           / 'timeseries_individual_buildings' / self.census_regions[int(ts_idx[1])]
                           / 'upgrade=0' / f'puma={ts_idx[2]}'), columns=['timestamp', bldg_id])

        # Order by timestamp
        df = df.to_pandas().sort_values(by='timestamp')
        # Slice each column from seq_ptr-context_len : seq_ptr + pred_len
        time_features = self.time_transform.transform(df['timestamp'].iloc[seq_ptr-self.context_len : seq_ptr+self.pred_len ])
        # Running faiss on CPU is  slower than on GPU, so we quantize loads later after data loading.
        load_features = df[bldg_id].iloc[seq_ptr-self.context_len : seq_ptr+self.pred_len ].values.astype(np.float32)  # (context_len+pred_len,)
        # For BoxCox transform
        if self.apply_scaler_transform != '':
            load_features = self.load_transform.transform(load_features)
        latlon_features = self.spatial_transform.transform(ts_idx[2]).repeat(self.context_len + self.pred_len, axis=0) 
        
        # residential = 0 and commercial = 1
        building_features = np.ones((self.context_len + self.pred_len,1), dtype=np.int32) * int(int(ts_idx[0]) % 2 == 0)

        sample = {
            'latitude': latlon_features[:, 0][...,None],
            'longitude': latlon_features[:, 1][...,None],
            'day_of_year': time_features[:, 0][...,None],
            'day_of_week': time_features[:, 1][...,None],
            'hour_of_day': time_features[:, 2][...,None],
            'building_type': building_features,
            'load': load_features[...,None]
        }
        return sample
    




import unittest
import numpy as np
import torch
from buildings_bench.evaluation.scoring_rules import ContinuousRankedProbabilityScore
from buildings_bench.evaluation.scoring_rules import RankedProbabilityScore


class TestRPS(unittest.TestCase):
    """
    Test the categorical ranked probability score
    """
    def setUp(self):
        np.random.seed(1984)

        
    def test_rps(self):
        rps = RankedProbabilityScore()
        y_true = torch.from_numpy(np.random.randint(0, 10, size=(2, 3, 1)))
        y_pred_logits = torch.from_numpy(np.random.normal(size=(2, 3, 10)))    
        bin_values = torch.from_numpy(np.random.normal(size=(10,)))        
        rps(None, y_true, y_pred_logits, bin_values)
        
    def test_bin_widths(self):
        rps = RankedProbabilityScore()
        bin_values = torch.FloatTensor([1., 5., 7.])
        y_true = torch.FloatTensor([[[0.], [1.]]]) # batch_size 1, seq_len 2, 1
        y_pred_logits = torch.FloatTensor([[[0.9, 0.1, 0.], [0.1, 0.9, 0.]]]) # batch_size 1, seq_len 2, vocab_size 3
        rps(None, y_true, y_pred_logits, bin_values)
        rps.mean()
        print(rps.value)


class TestContinuousRPS(unittest.TestCase):
    """
    https://github.com/properscoring/properscoring/blob/master/properscoring/tests/test_crps.py
    """

    def setUp(self):
        np.random.seed(1983)
        shape = (2, 3, 1)
        self.mu = torch.from_numpy(np.random.normal(size=shape))
        self.sig = torch.from_numpy(np.square(np.random.normal(size=shape)))
        self.params = torch.concatenate([self.mu, self.sig], dim=-1)

        self.obs = torch.from_numpy(np.random.normal(loc=self.mu, scale=self.sig, size=shape))
        self.crps = ContinuousRankedProbabilityScore()


    def test_continuous_rps(self):
        self.crps(self.obs, None, self.params, None)
        self.crps.mean()
        print(self.crps.value)


    def test_continuous_rps_correct(self):
        from properscoring import crps_ensemble
        from scipy import special

        n = 1000
        q = np.linspace(0. + 0.5 / n, 1. - 0.5 / n, n)
        # convert to the corresponding normal deviates
        normppf = special.ndtri
        z = normppf(q)

        sig = self.sig.squeeze(2).numpy()
        mu = self.mu.squeeze(2).numpy()
        forecasts = (z.reshape(-1, 1, 1) * sig) + mu
        expected = crps_ensemble(self.obs.squeeze(2).numpy(),
                                  forecasts, axis=0)
        expected = expected.mean(0)

        self.crps(self.obs, None, self.params, None)
        self.crps.mean()
        actual = self.crps.value

        np.testing.assert_allclose(actual, expected, rtol=1e-4)



import unittest
from buildings_bench import evaluation
import torch 


class TestMetrics(unittest.TestCase):
    def setUp(self):
        # Test torch tensor
        self.y_true = torch.FloatTensor([1, 2, 3]).view(1,3,1)
        self.y_pred = torch.FloatTensor([1, 2, 3]).view(1,3,1)

    def test_mae(self):
        mae = evaluation.metrics_factory('mae', 
                                         types=[evaluation.MetricType.SCALAR])
        self.assertEqual(len(mae), 1)

        mae = mae[0]
        self.assertEqual(mae.name, 'mae-scalar')
        self.assertEqual(mae.type, evaluation.MetricType.SCALAR)
        self.assertEqual(mae.UNUSED_FLAG, True)
        # Test call
        mae(self.y_true, self.y_pred)
        mae.mean()

        self.assertEqual(
            mae.value,
            torch.FloatTensor([0])
        )
        self.assertEqual(mae.UNUSED_FLAG, False)


    def test_cvrmse(self):
        cvrmse = evaluation.metrics_factory('cvrmse',
                                            types=[evaluation.MetricType.SCALAR])
        self.assertEqual(len(cvrmse), 1)
        cvrmse = cvrmse[0]
        self.assertEqual(cvrmse.name, 'cvrmse-scalar')
        self.assertEqual(cvrmse.type, evaluation.MetricType.SCALAR)
        self.assertEqual(cvrmse.UNUSED_FLAG, True)
        # Test call
        cvrmse(self.y_true, self.y_pred)
        cvrmse.mean()
        self.assertEqual(
            cvrmse.value,
            torch.FloatTensor([0])
        )
        self.assertEqual(cvrmse.UNUSED_FLAG, False)


    def test_mbe(self):
        mbe = evaluation.metrics_factory('mbe',
                                         types=[evaluation.MetricType.SCALAR])
        self.assertEqual(len(mbe), 1)
        mbe = mbe[0]
        self.assertEqual(mbe.name, 'mbe-scalar')
        self.assertEqual(mbe.type, evaluation.MetricType.SCALAR)
        self.assertEqual(mbe.UNUSED_FLAG, True)
        # Test call
        mbe(self.y_true, self.y_pred)
        mbe.mean()
        self.assertEqual(
            mbe.value,
            torch.FloatTensor([0])
        )
        self.assertEqual(mbe.UNUSED_FLAG, False)


import torch
from buildings_bench.models.base_model import BaseModel
import torch.nn as nn
import torch.nn.functional as F
import numpy as np


class moving_avg(nn.Module):
    """Moving average block to highlight the trend of time series
    """
    def __init__(self, kernel_size, stride):
        super(moving_avg, self).__init__()
        self.kernel_size = kernel_size
        self.avg = nn.AvgPool1d(kernel_size=kernel_size, stride=stride, padding=0)

    def forward(self, x):
        # padding on the both ends of time series
        front = x[:, 0:1, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        end = x[:, -1:, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        x = torch.cat([front, x, end], dim=1)
        x = self.avg(x.permute(0, 2, 1))
        x = x.permute(0, 2, 1)
        return x


class series_decomp(nn.Module):
    """Series decomposition block
    """
    def __init__(self, kernel_size):
        super(series_decomp, self).__init__()
        self.moving_avg = moving_avg(kernel_size, stride=1)

    def forward(self, x):
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean

class DLinearRegression(BaseModel):
    """
    Decomposition-Linear
    """
    def __init__(self,context_len=168, pred_len=24, continuous_loads=True):
        super(DLinearRegression, self).__init__(context_len, pred_len, continuous_loads)

        # Decompsition Kernel Size
        kernel_size = 25
        self.decompsition = series_decomp(kernel_size)
        # self.individual = True
        # self.channels = 1
       
        self.Linear_Seasonal = nn.Linear(context_len, self.pred_len)
        self.Linear_Trend = nn.Linear(context_len, self.pred_len)
        
            # Use this two lines if you want to visualize the weights
            # self.Linear_Seasonal.weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))
            # self.Linear_Trend.weight = nn.Parameter((1/self.seq_len)*torch.ones([self.pred_len,self.seq_len]))

    def forward(self, x):
        # x: [Batch, Input length, Channel]
        src_series = x['load'][:,:self.context_len,:]
        seasonal_init, trend_init = self.decompsition(src_series)
        seasonal_init, trend_init = seasonal_init.permute(0,2,1), trend_init.permute(0,2,1)
    
        seasonal_output = self.Linear_Seasonal(seasonal_init)
        trend_output = self.Linear_Trend(trend_init)

        x = seasonal_output + trend_output
        return x.permute(0,2,1) # to [Batch, Output length, Channel]
    
    def loss(self, x, y):
        return torch.nn.functional.mse_loss(x,y)
    
    def predict(self, x):
        return self.forward(x), None

    def unfreeze_and_get_parameters_for_finetuning(self):
        for p in self.parameters():
            p.requires_grad = True
        return self.parameters()
    
    def load_from_checkpoint(self, checkpoint_path):
        return None



import unittest
from buildings_bench.evaluation.managers import MetricsManager, DatasetMetricsManager
from buildings_bench.evaluation import metrics_factory
from buildings_bench.evaluation import scoring_rule_factory
from buildings_bench import BuildingTypes
import torch


class TestMetricsManager(unittest.TestCase):
       
        
    def test_create_dataset_metrics_manager_single_metric(self):
        metrics_manager = MetricsManager(
            metrics=metrics_factory('mae')
        )

        self.assertEqual(len(metrics_manager.metrics[BuildingTypes.RESIDENTIAL]), 1)
        self.assertEqual(len(metrics_manager.metrics[BuildingTypes.COMMERCIAL]), 1)
        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].name, 'mae-scalar')

    def test_create_dataset_metrics_manager_multiple_metrics(self):
        metrics_manager = MetricsManager(
            metrics=metrics_factory('mae') + metrics_factory('rmse')
        )

        self.assertEqual(len(metrics_manager.metrics[BuildingTypes.RESIDENTIAL]), 2)
        self.assertEqual(len(metrics_manager.metrics[BuildingTypes.COMMERCIAL]), 2)
        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].name, 'mae-scalar')
        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][1].name, 'rmse-scalar')
    
    def test_create_dataset_metrics_manager_with_scoring_rule(self):
        metrics_manager = MetricsManager(
            scoring_rule=scoring_rule_factory('rps')
        )

        self.assertEqual(len(metrics_manager.scoring_rules), 2)
        self.assertEqual(metrics_manager.scoring_rules[BuildingTypes.RESIDENTIAL].name, 'rps')
        self.assertEqual(metrics_manager.scoring_rules[BuildingTypes.COMMERCIAL].name, 'rps')

    def test_create_dataset_metrics_manager_with_scoring_rule_and_metrics(self):
        metrics_manager = MetricsManager(
            metrics=metrics_factory('mae'),
            scoring_rule=scoring_rule_factory('rps')
        )

        self.assertEqual(len(metrics_manager.metrics[BuildingTypes.RESIDENTIAL]), 1)
        self.assertEqual(len(metrics_manager.metrics[BuildingTypes.COMMERCIAL]), 1)
        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].name, 'mae-scalar')
        self.assertEqual(metrics_manager.scoring_rules[BuildingTypes.RESIDENTIAL].name, 'rps')
        self.assertEqual(metrics_manager.scoring_rules[BuildingTypes.COMMERCIAL].name, 'rps')

    def test_call_dataset_metrics_manager(self):

        metrics_manager = MetricsManager(
            metrics=metrics_factory('mae')
        )

        metrics_manager(
            y_true=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            y_pred=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            building_types_mask=torch.BoolTensor([False])
        )

        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].name, 'mae-scalar')
        metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].mean()
        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].value, 0)

    def test_update_loss_and_get_ppl_from_manager(self):

        metrics_manager = MetricsManager(
            metrics=metrics_factory('mae')
        )

        loss = torch.FloatTensor([1.0])
        metrics_manager(
            y_true=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            y_pred=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            building_types_mask=torch.BoolTensor([False]),
            loss=loss,
        )

        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].name, 'mae-scalar')
        metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].mean()
        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].value, 0)
        self.assertEqual(metrics_manager.get_ppl(), torch.exp( loss * 3 ) / 3)


    def test_summary(self):

        metrics_manager = MetricsManager(
            metrics=metrics_factory('mae')
        )

        loss = torch.FloatTensor([1.0])
        metrics_manager(
            y_true=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            y_pred=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            building_types_mask=torch.BoolTensor([False]),
            loss = loss
        )

        summary = metrics_manager.summary(with_loss=True, with_ppl=True)

        self.assertEqual(summary[BuildingTypes.RESIDENTIAL]['mae-scalar'].value, 0)
        self.assertEqual(summary['ppl'], torch.exp( loss * 3 ) / 3)
        self.assertEqual(summary['loss'], loss)

    def test_reset(self):

        metrics_manager = MetricsManager(
            metrics=metrics_factory('mae')
        )

        loss = torch.FloatTensor([1.0])
        metrics_manager(
            y_true=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            y_pred=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            building_types_mask=torch.BoolTensor([False]),
            loss = loss
        )

        metrics_manager.reset()

        self.assertEqual(len(metrics_manager.metrics[BuildingTypes.RESIDENTIAL]), 1)
        self.assertEqual(len(metrics_manager.metrics[BuildingTypes.COMMERCIAL]), 1)
        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].name, 'mae-scalar')
        self.assertEqual(metrics_manager.accumulated_unnormalized_loss, 0)

        metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].mean()
        metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].mean()

        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].UNUSED_FLAG, True)
        

    def test_building_type(self):

        metrics_manager = MetricsManager(
            metrics=metrics_factory('mae')
        )

        metrics_manager(
            y_true=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            y_pred=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            building_type = BuildingTypes.RESIDENTIAL_INT
        )

        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].name, 'mae-scalar')
        metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].mean()
        self.assertEqual(metrics_manager.metrics[BuildingTypes.RESIDENTIAL][0].value, 0)


class TestBenchmarkMetricsManager(unittest.TestCase):

    def test_create_benchmark_metrics_manager(self):
        metrics_manager = DatasetMetricsManager()

    def test_add_building_to_dataset_if_missing(self):
        metrics_manager = DatasetMetricsManager()

        metrics_manager.add_building_to_dataset_if_missing(
            dataset_name='test',
            building_id='0001')
        
        buliding_mm = metrics_manager.get_building_from_dataset(
            dataset_name='test',
            building_id='0001')
        
        # assert building_mm type is MetricsManager
        self.assertEqual(type(buliding_mm), MetricsManager)


    def test_call_benchmark_metrics_manager(self):

        metrics_manager = DatasetMetricsManager()

        metrics_manager.add_building_to_dataset_if_missing(
            dataset_name='test',
            building_id='0001')
        
        metrics_manager(
            dataset_name='test',
            building_id='0001',
            y_true=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            y_pred=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            building_types_mask=torch.BoolTensor([False])
        )

        buliding_mm = metrics_manager.get_building_from_dataset(
            dataset_name='test',
            building_id='0001')
        
        self.assertEqual(buliding_mm.metrics[BuildingTypes.RESIDENTIAL][0].name, 'cvrmse-scalar')
        buliding_mm.metrics[BuildingTypes.RESIDENTIAL][0].mean()
        self.assertEqual(buliding_mm.metrics[BuildingTypes.RESIDENTIAL][0].value, 0)

    def test_update_loss_and_get_ppl_from_benchmark_manager(self):

        metrics_manager = DatasetMetricsManager()

        metrics_manager.add_building_to_dataset_if_missing(
            dataset_name='test',
            building_id='0001')
        
        loss = torch.FloatTensor([1.0])
        metrics_manager(
            dataset_name='test',
            building_id='0001',
            y_true=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            y_pred=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            building_types_mask=torch.BoolTensor([False]),
            loss=loss,
        )

        buliding_mm = metrics_manager.get_building_from_dataset(
            dataset_name='test',
            building_id='0001')

        self.assertEqual(buliding_mm.get_ppl(), torch.exp( loss * 3 ) / 3)

    def test_benchmark_manager_summary(self):
        import pandas as pd

        metrics_manager = DatasetMetricsManager()

        metrics_manager.add_building_to_dataset_if_missing(
            dataset_name='test',
            building_id='0001')
        
        metrics_manager(
            dataset_name='test',
            building_id='0001',
            y_true=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            y_pred=torch.FloatTensor([1, 2, 3]).view(1,3,1),
            building_types_mask=torch.BoolTensor([False]),
        )

        summary = metrics_manager.summary()

        print(summary)



import faiss
import faiss.contrib.torch_utils 
from typing import Union
import numpy as np
from pathlib import Path
import torch


class LoadQuantizer:
    """Quantize load timeseries with KMeans. Merge centroids that are within a threshold."""
    def __init__(self, seed: int = 1, num_centroids: int = 2274,
                       with_merge=False, merge_threshold=0.01, device: str = 'cpu'):
        """
        Args:
            seed (int): random seed. Default: 1.
            num_centroids (int): number of centroids: Default: 2274.
            with_merge (bool): whether to merge centroids that are within a threshold: Default: False.
            merge_threshold (float): threshold for merging centroids. Default: 0.01 (kWh).
            device (str): cpu or cuda. Default: cpu.
        """
        self.seed = seed
        self.K = num_centroids
        self.with_merge = with_merge
        self.merge_threshold = merge_threshold
        self.kmeans = None
        self.original_order_to_sorted_centroid_indices = None
        self.original_order_to_merged_centroid_map = None
        self.merged_centroids = None
        self.device = device

    def get_vocab_size(self) -> int:
        if self.with_merge:
            return len(self.merged_centroids)
        return self.K

    def train(self, sample: np.ndarray) -> None:
        """Fit KMeans to a subset of the data. 
        
        Optionally, merge centroids that are within a threshold.

        Args:
            sample (np.ndarray): shape [num_samples, 1]
        """
        self.kmeans = faiss.Kmeans(d=1, k=self.K, niter=300, nredo=10,
                     seed=self.seed, verbose=True, gpu=(True if 'cuda' in self.device else False))
        # max is 256, min is 39 per centroid by defaul
        # https://github.com/facebookresearch/faiss/blob/d5ca6f0a79aa382bb98d2221a8f61c1a200efa25/faiss/Clustering.cpp#L2t
        self.kmeans.train(sample.reshape(-1,1))
        
        if self.with_merge:
            # sort the centroids and return the sorted indices
            sorted_indices = np.argsort(self.kmeans.centroids.reshape(-1))
            # sort the centroids
            #unsorted_index_to_value_map = self.kmeans.centroids.copy()
            index_to_value_map = self.kmeans.centroids[sorted_indices]
            # create a map from the sorted indices to the original indices
            self.original_order_to_sorted_centroid_indices = np.zeros_like(sorted_indices)
            for i in range(len(sorted_indices)):
                self.original_order_to_sorted_centroid_indices[sorted_indices[i]] = i

            # Iterate over sorted centroids, merge centroids that are within a threshold
            previous_centroid_value = index_to_value_map[0]
            merge_classes = [0]
            candidate_centroids = [previous_centroid_value]

            # new list of centroids
            self.merged_centroids = []
            # list that maps the original centroid indices to the new centroid indices
            candidate_index_list = []


            #  Iterate over the sorted centroids.
            for i in range(1,len(index_to_value_map)):
                # If the current centroid is within a threshold of the previous centroid,
                # add it to the merged class. 
                if index_to_value_map[i] - previous_centroid_value < self.merge_threshold:
                    merge_classes += [i]
                    candidate_centroids += [index_to_value_map[i]]
                #  If the current centroid is not within a threshold of the previous centroid,
                #  merge the set of candidate centroids, add the merged class indices to the
                # candidate list, and update previous centroid value to point to current index
                else:
                    # I take the average of the centroids
                    # being merged
                    self.merged_centroids += [np.mean(candidate_centroids)]
                    candidate_index_list += [merge_classes]
                    # reset this with current index
                    merge_classes = [i]
                    candidate_centroids = [index_to_value_map[i]]
                    previous_centroid_value = index_to_value_map[i]
            # Add the last centroid to the new list of centroids and add the merged class to the merge list.
            self.merged_centroids += [np.mean(candidate_centroids)]
            candidate_index_list += [merge_classes]
            self.merged_centroids = np.array(self.merged_centroids)
            # Create a map from the original centroid index to the new centroid index
            self.original_order_to_merged_centroid_map = np.zeros(len(index_to_value_map), dtype=np.int32)
            for i in range(len(candidate_index_list)):
                for j in candidate_index_list[i]:
                    self.original_order_to_merged_centroid_map[j] = i
            #print('Merged {} centroids to {} centroids'.format(len(index_to_value_map), len(self.merged_centroids)))
            self.K = len(self.merged_centroids)
            
    def save(self, output_path: Path) -> None:
        if 'cuda' in self.device:
            self.kmeans.index = faiss.index_gpu_to_cpu(self.kmeans.index)

        chunk = faiss.serialize_index(self.kmeans.index)
        np.save(output_path / "kmeans_K={}.npy".format(self.K), chunk)
        np.save(output_path / "kmeans_centroids_K={}.npy".format(self.K), self.kmeans.centroids)
        if self.with_merge:
            np.save(output_path / "kmeans_original_to_sorted_indices_K={}.npy".format(self.K),
                    self.original_order_to_sorted_centroid_indices)
            np.save(output_path / "kmeans_original_to_merged_map_K={}.npy".format(self.K),
                    self.original_order_to_merged_centroid_map)
            np.save(output_path / "kmeans_merged_centers_K={}.npy".format(self.K),
                    self.merged_centroids)

    def load(self, saved_path: Path) -> None:
        chunk = np.load(
            saved_path / "kmeans_K={}.npy".format(self.K), allow_pickle=True)
        self.kmeans = faiss.Kmeans(d=1, k=self.K, niter=200, nredo=20,
                     seed=self.seed, verbose=True, gpu=(True if 'cuda' in self.device else False))
        self.kmeans.index = faiss.deserialize_index(chunk)
        self.kmeans.centroids = np.load(
            saved_path / "kmeans_centroids_K={}.npy".format(self.K), allow_pickle=True)

        if self.with_merge:
            self.original_order_to_sorted_centroid_indices = np.load(
                saved_path / "kmeans_original_to_sorted_indices_K={}.npy".format(self.K), allow_pickle=True)
            self.original_order_to_merged_centroid_map = np.load(
                saved_path / "kmeans_original_to_merged_map_K={}.npy".format(self.K), allow_pickle=True)
            self.merged_centroids = np.load(
                saved_path / "kmeans_merged_centers_K={}.npy".format(self.K), allow_pickle=True)
            
            #print(f'Loaded Kmeans quantizer with K={len(self.merged_centroids)}')
        #else:
        #print(f'Loaded Kmeans quantizer with K={self.K}')

        if 'cuda' in self.device:
            local_rank = int(self.device.split(':')[-1])
            faiss_res = faiss.StandardGpuResources()
            self.kmeans.index = faiss.index_cpu_to_gpu(faiss_res, local_rank, self.kmeans.index)
            self.kmeans.centroids = torch.from_numpy(self.kmeans.centroids).float().to(self.device)
            self.kmeans.centroids.squeeze()
            if self.with_merge:
                self.original_order_to_sorted_centroid_indices = torch.from_numpy( 
                    self.original_order_to_sorted_centroid_indices).long().to(self.device)
                self.original_order_to_merged_centroid_map = torch.from_numpy(
                    self.original_order_to_merged_centroid_map).long().to(self.device)
                self.merged_centroids = torch.from_numpy(
                    self.merged_centroids).float().to(self.device)
                self.merged_centroids.squeeze()
            #print(f'Kmeans quantizer moved to GPU: {type(self.kmeans.index)}')


    def transform(self, sample: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
        """Quantize a sample of load values into a sequence of indices.
        
        Args:
            sample Union[np.ndarray, torch.Tensor]: of shape (n, 1) or (b,n,1). 
                type is numpy if device is cpu or torch Tensor if device is cuda.

        Returns:
            sample (Union[np.ndarray, torch.Tensor]): of shape (n, 1) or (b,n,1).
        """
        init_shape = sample.shape
        sample = sample.reshape(-1,1)
        sample = self.kmeans.index.search(sample, 1)[1].reshape(-1)
        if self.with_merge:
            sample = self.original_order_to_sorted_centroid_indices[sample]
            sample = self.original_order_to_merged_centroid_map[sample]
        sample = sample.reshape(init_shape)
        return sample


    def undo_transform(self, sample: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
        """Dequantize a sample of integer indices into a sequence of load values.

        Args:
            sample Union[np.ndarray, torch.Tensor]: of shape (n, 1) or (b,n,1). 
                type is numpy if device is cpu or torch Tensor if device is cuda.

        Returns:
            sample (Union[np.ndarray, torch.Tensor]): of shape (n, 1) or (b,n,1).
        """
        init_shape = sample.shape
        sample = sample.reshape(-1)
        if self.with_merge:
            sample = self.merged_centroids[sample]
        else:
            sample = self.kmeans.centroids[sample]
        sample = sample.reshape(init_shape)
        return sample



import torch
from buildings_bench.models.base_model import BaseModel
from buildings_bench.models.transformers import TimeSeriesSinusoidalPeriodicEmbedding
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from typing import Tuple, Dict


class DeepAutoregressiveRNN(BaseModel):

    def __init__(self, hidden_size = 40, lstm_layers = 3, 
                 context_len = 168, pred_len=24,
                 for_pretraining=False, 
                 continuous_head='gaussian_nll',
                 continuous_loads=True):
        super(DeepAutoregressiveRNN,self).__init__(context_len, pred_len, continuous_loads)
        self.continuous_head = continuous_head
        out_dim = 1 if self.continuous_head == 'mse' else 2
        self.logits = nn.Linear(hidden_size, out_dim)
        
        self.for_pretraining = for_pretraining
        if self.for_pretraining:

            self.power_embedding = nn.Linear(1, 64)
            self.building_embedding = nn.Embedding(2, 32)
            self.lat_embedding = nn.Linear(1, 32)
            self.lon_embedding = nn.Linear(1, 32)
            self.day_of_year_encoding = TimeSeriesSinusoidalPeriodicEmbedding(32) 
            self.day_of_week_encoding = TimeSeriesSinusoidalPeriodicEmbedding(32)
            self.hour_of_day_encoding = TimeSeriesSinusoidalPeriodicEmbedding(32)

            self.encoder = nn.LSTM(256, hidden_size, num_layers=lstm_layers, batch_first=True)
            self.decoder = nn.LSTM(256, hidden_size, num_layers=lstm_layers, batch_first=True)
        else:
            self.encoder = nn.LSTM(1, hidden_size, num_layers=lstm_layers, batch_first=True)
            self.decoder = nn.LSTM(1, hidden_size, num_layers=lstm_layers, batch_first=True)

    def forward(self, x):
        r"""Forward pass of the deep autoregressive model. 

        Args:
            x (Dict): dictionary of input tensors.
        Returns:
            logits (torch.Tensor): [batch_size, pred_len, vocab_size] if not continuous_loads,
                                   [batch_size, pred_len, 1] if continuous_loads and continuous_head == 'mse', 
                                   [batch_size, pred_len, 2] if continuous_loads and continuous_head == 'gaussian_nll'.
        """
        if self.for_pretraining:
            # [batch_size, seq_len, 256]
            time_series_embed = torch.cat([
                self.lat_embedding(x['latitude']),
                self.lon_embedding(x['longitude']),
                self.building_embedding(x['building_type']).squeeze(2),
                self.day_of_year_encoding(x['day_of_year']),
                self.day_of_week_encoding(x['day_of_week']),
                self.hour_of_day_encoding(x['hour_of_day']),
                self.power_embedding(x['load']).squeeze(2),
            ], dim=2)
        else:
            time_series_embed = x['load']

        # [batch_size, context_len, d]
        src_series_inputs = time_series_embed[:, :self.context_len, :]
        # [batch_size, pred_len, d]
        # The last element of the target sequence is not used as input
        # The last element of the source sequence is used as the initial decoder input
        tgt_series_inputs = time_series_embed[:, self.context_len-1 : -1, :]
        _, (h_n, c_n) = self.encoder(src_series_inputs)
        outs, _ = self.decoder(tgt_series_inputs, (h_n, c_n))
        return self.logits(outs)

    def predict(self, x: Dict) -> Tuple[torch.Tensor, torch.Tensor]:
        return self.generate_sample(x, greedy=True)
    
    def loss(self, x, y):
        if self.continuous_head == 'mse':
            return F.mse_loss(x, y)
        elif self.continuous_head == 'gaussian_nll':
            return F.gaussian_nll_loss(x[:, :, 0].unsqueeze(2), y,
                                       F.softplus(x[:, :, 1].unsqueeze(2)) **2)

    def unfreeze_and_get_parameters_for_finetuning(self):
        return self.parameters()

    def load_from_checkpoint(self, checkpoint_path):
        stored_ckpt = torch.load(checkpoint_path)
        model_state_dict = stored_ckpt['model']
        new_state_dict = {}
        for k,v in model_state_dict.items():
            # remove string 'module.' from the key
            if 'module.' in k:
                new_state_dict[k.replace('module.', '')] = v
            else:
                new_state_dict[k] = v
        self.load_state_dict(new_state_dict)    
        #print(f"Loaded model checkpoint from {checkpoint_path}...")


    @torch.no_grad()
    def generate_sample(self, 
                 x,
                 temperature=1.0,
                 greedy=False,
                 num_samples=1):
        """Sample from the conditional distribution.

        Use output of decoder at each prediction step as input to the next decoder step.
        Implements greedy decoding and random temperature-controlled sampling.
        
        Args:
            x (Dict): dictionary of input tensors
            temperature (float): temperature for sampling
            greedy (bool): whether to use greedy decoding
            num_samples (int): number of samples to generate
        
        Returns:
            predictions (torch.Tensor): of shape [batch_size, pred_len, 1] or shape [batch_size, num_samples, pred_len] if num_samples > 1.
            distribution_parameters (torch.Tensor): of shape [batch_size, pred_len, 1]. Not returned if sampling.
        """
        if self.for_pretraining:
            time_series_embed = torch.cat([
                self.lat_embedding(x['latitude']),
                self.lon_embedding(x['longitude']),
                self.building_embedding(x['building_type']).squeeze(2),
                self.day_of_year_encoding(x['day_of_year']),
                self.day_of_week_encoding(x['day_of_week']),
                self.hour_of_day_encoding(x['hour_of_day']),
                self.power_embedding(x['load']).squeeze(2),
            ], dim=2)
        else:
            time_series_embed = x['load']
        # [batch_size, context_len, d_model]
        src_series_inputs = time_series_embed[:, :self.context_len, :]
        tgt_series_inputs = time_series_embed[:, self.context_len-1 : -1, :]

        _, (h_n, c_n) = self.encoder(src_series_inputs)
        decoder_input = tgt_series_inputs[:, 0, :].unsqueeze(1)
        if num_samples > 1 and not greedy:
            # [batch_size, 1, emb_size] --> [batch_size * num_sampes, 1, emb_size]
            decoder_input = decoder_input.repeat_interleave(num_samples, dim=0)
            # [num_layers, batch_size, hidden_size] -> [num_layers, batch_size * num_samples, hidden_size]
            h_n = h_n.repeat_interleave(num_samples, dim=1)
            c_n = c_n.repeat_interleave(num_samples, dim=1)
        all_preds, all_logits = [], []
        for k in range(1, self.pred_len+1):
            decoder_output, (h_n, c_n) = self.decoder(decoder_input, (h_n, c_n))
            # [batch_size, 1] if continuous (2 if head is gaussian_nll) or [batch_size, vocab_size] if not continuous_loads
            outputs = self.logits(decoder_output[:, -1, :])
            all_logits += [outputs.unsqueeze(1)]

            if self.continuous_head == 'mse':
                all_preds += [outputs] 
            elif self.continuous_head == 'gaussian_nll':
                if greedy:
                    all_preds += [outputs[:, 0].unsqueeze(1)] # mean only
                    outputs = all_preds[-1] # [batch_size, 1, 1]
                else:
                    mean = outputs[:,0]
                    std= torch.nn.functional.softplus(outputs[:,1])
                    outputs = torch.distributions.normal.Normal(mean, std).sample().unsqueeze(1)
                    all_preds += [outputs]    
                
            # [batch_size, d_model]
            if k < self.pred_len:
                # [batch_size, d_model]
                next_decoder_input = tgt_series_inputs[:, k]
                if num_samples > 1 and not greedy:
                    # [batch_size, d_model] --> [batch_size * num_samples, d_model]
                    next_decoder_input = next_decoder_input.repeat_interleave(num_samples, dim=0)
                if self.for_pretraining:
                    # Use the embedding predicted load instead of the ground truth load
                    embedded_pred = self.power_embedding(outputs)  
                    next_decoder_input = torch.cat([ next_decoder_input[:, :-embedded_pred.shape[-1]], embedded_pred ], dim=1)
                else:
                    next_decoder_input = torch.cat([ next_decoder_input[:, :-1], outputs ], dim=1)
                # Append the next decoder input to the decoder input
                decoder_input = torch.cat([decoder_input, next_decoder_input.unsqueeze(1)], dim=1)
        if num_samples == 1 or greedy:
            if self.continuous_head == 'gaussian_nll':
                # [batch_size, pred_len, 2]
                gaussian_params = torch.stack(all_logits,1)[:,:,0,:]
                means = gaussian_params[:,:,0]
                sigma = torch.nn.functional.softplus(gaussian_params[:,:,1])
                return torch.stack(all_preds,1), torch.cat([means.unsqueeze(2), sigma.unsqueeze(2)],2)
            else:
                return torch.stack(all_preds,1), torch.stack(all_logits,1)[:,:,0,:]
        else:
            # [batch_size, num_samples, pred_len]
            return torch.stack(all_preds,1).reshape(-1, num_samples, self.pred_len)        


import math
import torch
from typing import Tuple, Dict
from torch import nn
import torch.nn.functional as F
from torch.nn import Transformer
import numpy as np
from buildings_bench.models.base_model import BaseModel


class TokenEmbedding(nn.Module):
    """Helper Module to convert tensor of input
       indices into corresponding tensor of token embeddings.
    """
    def __init__(self, vocab_size: int, emb_size: int):
        """
        Args:
            vocab_size (int): number of quantized load values in the entire vocabulary.
            emb_size (int): embedding size.
        """
        super(TokenEmbedding, self).__init__()
        self.embedding = nn.Embedding(vocab_size, emb_size)
        self.emb_size = emb_size

    def forward(self, tokens: torch.Tensor):
        return self.embedding(tokens.long()) * math.sqrt(self.emb_size)


class PositionalEncoding(nn.Module):
    """Helper Module that adds positional encoding to the token embedding to
       introduce a notion of order within a time-series.
    """
    def __init__(self,
                 emb_size: int,
                 dropout: float,
                 maxlen: int = 500):
        """
        Args:
            emb_size (int): embedding size.
            dropout (float): dropout rate.
            maxlen (int): maximum possible length of the incoming time series.
        """
        super(PositionalEncoding, self).__init__()
        den = torch.exp(- torch.arange(0, emb_size, 2)* math.log(10000) / emb_size)
        pos = torch.arange(0, maxlen).reshape(maxlen, 1)
        pos_embedding = torch.zeros((maxlen, emb_size))
        pos_embedding[:, 0::2] = torch.sin(pos * den)
        pos_embedding[:, 1::2] = torch.cos(pos * den)
        pos_embedding = pos_embedding.unsqueeze(-2)

        self.dropout = nn.Dropout(dropout)
        self.register_buffer('pos_embedding', pos_embedding)

    def forward(self, token_embedding: torch.Tensor) -> torch.Tensor:
        # batch first - use size(1)
        # need to permute token embeddings from [batch_size, seqlen x emb_size] to [seqlen x batch_size, emb_size]
        return self.dropout(token_embedding.permute(1,0,2) + self.pos_embedding[:token_embedding.size(1), :]).permute(1,0,2)


class TimeSeriesSinusoidalPeriodicEmbedding(nn.Module):
    """This module produces a sinusoidal periodic embedding for a sequence of values in [-1, +1]."""
    def __init__(self, embedding_dim: int) -> None:
        """
        Args:
            embedding_dim (int): embedding size.
        """
        super().__init__()
        self.linear = nn.Linear(2, embedding_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """`x` is expected to be [batch_size, seqlen, 1]."""
        with torch.no_grad():
            x = torch.cat([torch.sin(np.pi * x), torch.cos(np.pi * x)], dim=2)
        # [batch_size, seqlen x 2] --> [batch_size, seqlen, embedding_dim]
        return self.linear(x)


class ZeroEmbedding(nn.Module):
    """ Outputs zeros of the desired output dim."""
    def __init__(self, embedding_dim: int): 
        """
        Args:
            embedding_dim (int): embedding size.
        """
        super().__init__()
        self.embedding_dim = embedding_dim
        self.zeros_embedding = nn.Parameter(
            torch.zeros(1, 1, embedding_dim), requires_grad=False)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """ `x` is expected to be [batch_size, seqlen, 1]."""
        return self.zeros_embedding.repeat(x.shape[0], x.shape[1], 1)
    

class LoadForecastingTransformer(BaseModel):
    """
    An encoder-decoder time series Transformer. Based on PyTorch nn.Transformer.

    - Uses masking in the decoder to prevent the model from peeking into the future
    - Uses N(0, 0.02) for weight initialization
    - Trains with teacher forcing (i.e. the target is used as the input to the decoder)
    - continuous_loads (True) just predict target values
                     (False) categorical over quantized load values
    """
    def __init__(self,
                 context_len: int = 168,
                 pred_len: int = 24,
                 vocab_size = 2274,
                 num_encoder_layers: int = 3,
                 num_decoder_layers: int = 3,
                 d_model: int = 256,
                 nhead: int = 8,
                 dim_feedforward: int = 256,
                 dropout: float = 0.0,
                 activation: str = 'gelu',
                 continuous_loads = False,
                 continuous_head = 'mse',
                 ignore_spatial = False):
        """
        Args:
            context_len (int): length of the input sequence.
            pred_len (int): length of the output sequence.
            vocab_size (int): number of quantized load values in the entire vocabulary.
            num_encoder_layers (int): number of encoder layers.
            num_decoder_layers (int): number of decoder layers.
            d_model (int): number of expected features in the encoder/decoder inputs.
            nhead (int): number of heads in the multi-head attention models.
            dim_feedforward (int): dimension of the feedforward network model.
            dropout (float): dropout value.
            activation (str): the activation function of encoder/decoder intermediate layer, relu or gelu.
            continuous_loads (bool): whether inputs are continuous/to train the model to predict continuous values.
            continuous_head (str): 'mse' or 'gaussian_nll'.
            ignore_spatial (bool): whether to ignore the spatial features.
        """
        super().__init__(context_len, pred_len, continuous_loads)
        
        self.continuous_head = continuous_head
        self.vocab_size = vocab_size
        self.ignore_spatial = ignore_spatial
        s = d_model // 256

        self.transformer = Transformer(d_model=d_model,
                                       nhead=nhead,
                                       num_encoder_layers=num_encoder_layers,
                                       num_decoder_layers=num_decoder_layers,
                                       dim_feedforward=dim_feedforward,
                                       dropout=dropout,
                                       activation=activation,
                                       batch_first=True)

        if self.continuous_loads:
            out_dim = 1 if self.continuous_head == 'mse' else 2
            self.logits = nn.Linear(d_model, out_dim)
            self.power_embedding = nn.Linear(1, 64 * s)
        else:
            self.logits = nn.Linear(d_model, self.vocab_size)
            self.power_embedding = TokenEmbedding(self.vocab_size, 64 * s)
        self.tgt_mask = self.transformer.generate_square_subsequent_mask(self.pred_len)
        self.positional_encoding = PositionalEncoding(
            d_model, dropout=dropout)
        self.building_embedding = nn.Embedding(2, 32 * s)
        self.lat_embedding = nn.Linear(1, 32 * s)
        self.lon_embedding = nn.Linear(1, 32 * s)
        if self.ignore_spatial:
            self.lat_embedding = ZeroEmbedding(32 * s)
            self.lon_embedding = ZeroEmbedding(32 * s)
        self.day_of_year_encoding = TimeSeriesSinusoidalPeriodicEmbedding(32 * s) 
        self.day_of_week_encoding = TimeSeriesSinusoidalPeriodicEmbedding(32 * s)
        self.hour_of_day_encoding = TimeSeriesSinusoidalPeriodicEmbedding(32 * s)
        
    def to(self, device):
        self.tgt_mask = self.tgt_mask.to(device)
        return super().to(device)

    def forward(self, x):
        r"""Forward pass of the time series transformer. 

        Args:
            x (Dict): dictionary of input tensors.
        Returns:
            logits (torch.Tensor): [batch_size, pred_len, vocab_size] if not continuous_loads,
                                   [batch_size, pred_len, 1] if continuous_loads and continuous_head == 'mse', 
                                   [batch_size, pred_len, 2] if continuous_loads and continuous_head == 'gaussian_nll'.
        """
        # [batch_size, seq_len, d_model]
        time_series_embed = torch.cat([
            self.lat_embedding(x['latitude']),
            self.lon_embedding(x['longitude']),
            self.building_embedding(x['building_type']).squeeze(2),
            self.day_of_year_encoding(x['day_of_year']),
            self.day_of_week_encoding(x['day_of_week']),
            self.hour_of_day_encoding(x['hour_of_day']),
            self.power_embedding(x['load']).squeeze(2),
        ], dim=2)
        # [batch_size, context_len, d_model]
        src_series_inputs = time_series_embed[:, :self.context_len, :]
        # [batch_size, pred_len, d_model]
        # The last element of the target sequence is not used as input
        # The last element of the source sequence is used as the initial decoder input
        tgt_series_inputs = time_series_embed[:, self.context_len-1 : -1, :]
        src_series_embed = self.positional_encoding(src_series_inputs)
        tgt_series_embed = self.positional_encoding(tgt_series_inputs) 

        # The output of TransformerEncoder is the sequence from the last layer
        # The shape will be [batch_size, context_len, d_model]
        
        memory = self.transformer.encoder(src_series_embed, mask=None)
        outs = self.transformer.decoder(tgt_series_embed, memory, tgt_mask=self.tgt_mask)
        return self.logits(outs)

    def predict(self, x: Dict) -> Tuple[torch.Tensor, torch.Tensor]:
        return self.generate_sample(x, greedy=True)
    
    def loss(self, x, y):
        if self.continuous_loads and self.continuous_head == 'mse':
            return F.mse_loss(x, y)
        elif self.continuous_loads and self.continuous_head == 'gaussian_nll':
            return F.gaussian_nll_loss(x[:, :, 0].unsqueeze(2), y,
                                       F.softplus(x[:, :, 1].unsqueeze(2)) **2)
        else:
            return F.cross_entropy(x.reshape(-1, self.vocab_size),
                                             y.long().reshape(-1))
                    
    def unfreeze_and_get_parameters_for_finetuning(self):
        # for p in self.parameters():
        #     p.requires_grad_(False)
        # self.logits.requires_grad_(True)
        # return self.logits.parameters()
        return self.parameters()

    def load_from_checkpoint(self, checkpoint_path):
        stored_ckpt = torch.load(checkpoint_path)
        model_state_dict = stored_ckpt['model']
        new_state_dict = {}
        for k,v in model_state_dict.items():
            # remove string 'module.' from the key
            if 'module.' in k:
                new_state_dict[k.replace('module.', '')] = v
            else:
                new_state_dict[k] = v
        self.load_state_dict(new_state_dict)    
        #print(f"Loaded model checkpoint from {checkpoint_path}...")
    

    @torch.no_grad()
    def generate_sample(self, 
                 x,
                 temperature=1.0,
                 greedy=False,
                 num_samples=1):
        """Sample from the conditional distribution.

        Use output of decoder at each prediction step as input to the next decoder step.
        Implements greedy decoding and random temperature-controlled sampling.
        
        Top-k sampling and nucleus sampling are deprecated.

        Args:
            x (Dict): dictionary of input tensors
            temperature (float): temperature for sampling
            greedy (bool): whether to use greedy decoding
            num_samples (int): number of samples to generate
        
        Returns:
            predictions (torch.Tensor): of shape [batch_size, pred_len, 1] or shape [batch_size, num_samples, pred_len] if num_samples > 1.
            distribution_parameters (torch.Tensor): of shape [batch_size, pred_len, 1]. Not returned if sampling.
        """
        time_series_embed = torch.cat([
            self.lat_embedding(x['latitude']),
            self.lon_embedding(x['longitude']),
            self.building_embedding(x['building_type']).squeeze(2),
            self.day_of_year_encoding(x['day_of_year']),
            self.day_of_week_encoding(x['day_of_week']),
            self.hour_of_day_encoding(x['hour_of_day']),
            self.power_embedding(x['load']).squeeze(2),
        ], dim=2)
        # [batch_size, context_len, d_model]
        src_series_inputs = time_series_embed[:, :self.context_len, :]
        tgt_series_inputs = time_series_embed[:, self.context_len-1 : -1, :]
        src_series_embed = self.positional_encoding(src_series_inputs)

        encoder_output = self.transformer.encoder(src_series_embed)
        decoder_input = tgt_series_inputs[:, 0, :].unsqueeze(1)
        if num_samples > 1 and not greedy:
            # [batch_size, 1, emb_size] --> [batch_size * num_sampes, 1, emb_size]
            decoder_input = decoder_input.repeat_interleave(num_samples, dim=0)
            encoder_output = encoder_output.repeat_interleave(num_samples, dim=0)
        all_preds, all_logits = [], []
        for k in range(1, self.pred_len+1):
            decoder_embed = self.positional_encoding(decoder_input)
            tgt_mask = self.transformer.generate_square_subsequent_mask(k)
            decoder_output = self.transformer.decoder(decoder_embed, encoder_output, tgt_mask.to(encoder_output.device))
            # [batch_size, 1] if continuous (2 if head is gaussian_nll) or [batch_size, vocab_size] if not continuous_loads
            outputs = self.logits(decoder_output[:, -1, :])
            all_logits += [outputs.unsqueeze(1)]

            if self.continuous_loads:
                if self.continuous_head == 'mse':
                    all_preds += [outputs] 
                elif self.continuous_head == 'gaussian_nll':
                    if greedy:
                        all_preds += [outputs[:, 0].unsqueeze(1)] # mean only
                        outputs = all_preds[-1] # [batch_size, 1, 1]
                    else:
                        mean = outputs[:,0]
                        std= torch.nn.functional.softplus(outputs[:,1])
                        outputs = torch.distributions.normal.Normal(mean, std).sample().unsqueeze(1)
                        all_preds += [outputs]    

            elif not greedy:
                # Sample from a Categorical distribution with logits outputs
                all_preds += [torch.multinomial(torch.nn.functional.softmax(outputs/temperature, dim=1), 1)]
                # change outputs to the predicted load tokens
                outputs = all_preds[-1] # [batch_size * num_samples, 1]
            else:
                # outputs are [batch_size, vocab_size]
                # Greedy decoding
                all_preds += [outputs.argmax(dim=1).unsqueeze(1)]
                # change outputs to the predicted load tokens
                outputs = all_preds[-1]
                
            # [batch_size, d_model]
            if k < self.pred_len:
                # [batch_size, d_model]
                next_decoder_input = tgt_series_inputs[:, k]
                if num_samples > 1 and not greedy:
                    # [batch_size, d_model] --> [batch_size * num_samples, d_model]
                    next_decoder_input = next_decoder_input.repeat_interleave(num_samples, dim=0)
                # Use the embedding predicted load instead of the ground truth load
                embedded_pred = self.power_embedding(outputs)  
                if not self.continuous_loads:
                    # [batch_size, 1, 1, 64*s] --> [batch_size, 64*s]
                    embedded_pred = embedded_pred.squeeze(2).squeeze(1)
                next_decoder_input = torch.cat([ next_decoder_input[:, :-embedded_pred.shape[-1]], embedded_pred ], dim=1)
                # Append the next decoder input to the decoder input
                decoder_input = torch.cat([decoder_input, next_decoder_input.unsqueeze(1)], dim=1)
        if num_samples == 1 or greedy:
            if self.continuous_head == 'gaussian_nll':
                # [batch_size, pred_len, 2]
                gaussian_params = torch.stack(all_logits,1)[:,:,0,:]
                means = gaussian_params[:,:,0]
                sigma = torch.nn.functional.softplus(gaussian_params[:,:,1])
                return torch.stack(all_preds,1), torch.cat([means.unsqueeze(2), sigma.unsqueeze(2)],2)
            else:
                return torch.stack(all_preds,1), torch.stack(all_logits,1)[:,:,0,:]
        else:
            # [batch_size, num_samples, pred_len]
            return torch.stack(all_preds,1).reshape(-1, num_samples, self.pred_len)



import unittest
from buildings_bench import load_torch_dataset, load_pandas_dataset, load_pretraining
from buildings_bench.data.datasets import PandasBuildingDatasetsFromCSV
from buildings_bench import BuildingTypes
from pathlib import Path
import torch
import os

class TestLoadDatasets(unittest.TestCase):

    def test_dataset_not_in_registry(self):
        with self.assertRaises(ValueError):
            datasets = load_torch_dataset('not-a-dataset')


    def test_load_buildings900k_test(self):
        building_dataset_generator = load_torch_dataset('buildings-900k-test')
        for building_name, building in building_dataset_generator:
                print(f'building {building_name} dataset length: {len(building)}')
                # create a dataloader for the building
                building_dataloader = torch.utils.data.DataLoader(building, batch_size=256, shuffle=False)
                for sample in building_dataloader:
                    x = sample['load']
                    break
                break

    def test_load_electricity(self):
        building_dataset_generator = load_torch_dataset('electricity')
        for building_name, building in building_dataset_generator:
            print(f'building {building_name} dataset length: {len(building)}')
            # create a dataloader for the building
            building_dataloader = torch.utils.data.DataLoader(building, batch_size=256, shuffle=False)
            for sample in building_dataloader:
                x = sample['load']
                break
            break

    def test_load_sceaux(self):
        building_dataset_generator = load_torch_dataset('sceaux')
        for building_name, building in building_dataset_generator:
            print(f'building {building_name} dataset length: {len(building)}')
            # create a dataloader for the building
            building_dataloader = torch.utils.data.DataLoader(building, batch_size=256, shuffle=False)
            for sample in building_dataloader:
                x = sample['load']
                break
            break

    def test_load_borealis(self):
        building_dataset_generator = load_torch_dataset('borealis')
        for building_name, building in building_dataset_generator:
            print(f'building {building_name} dataset length: {len(building)}')
            # create a dataloader for the building
            building_dataloader = torch.utils.data.DataLoader(building, batch_size=256, shuffle=False)
            for sample in building_dataloader:
                x = sample['load']
                break
            break

    def test_load_bdg2_panther(self):
        building_dataset_generator = load_torch_dataset('bdg-2:panther')
        for building_name, building in building_dataset_generator:
            print(f'building {building_name} dataset length: {len(building)}')
            # create a dataloader for the building
            building_dataloader = torch.utils.data.DataLoader(building, batch_size=256, shuffle=False)
            for sample in building_dataloader:
                x = sample['load']
                break
            break

    def test_load_bdg2_rat(self):
        building_dataset_generator = load_torch_dataset('bdg-2:rat')
        for building_name, building in building_dataset_generator:
            print(f'building {building_name} dataset length: {len(building)}')
            # create a dataloader for the building
            building_dataloader = torch.utils.data.DataLoader(building, batch_size=256, shuffle=False)
            for sample in building_dataloader:
                x = sample['load']
                break
            break

    def test_load_ideal(self):
        building_dataset_generator = load_torch_dataset('ideal')
        for building_name, building in building_dataset_generator:
            print(f'building {building_name} dataset length: {len(building)}')
            # create a dataloader for the building
            building_dataloader = torch.utils.data.DataLoader(building, batch_size=256, shuffle=False)
            for sample in building_dataloader:
                x = sample['load']
                break
            break
    

    def test_load_lcl(self):
        building_dataset_generator = load_torch_dataset('lcl')
        for building_name, building in building_dataset_generator:
            print(f'building {building_name} dataset length: {len(building)}')
            # create a dataloader for the building
            building_dataloader = torch.utils.data.DataLoader(building, batch_size=256, shuffle=False)
            for sample in building_dataloader:
                x = sample['load']
                break
            break

    def test_load_smart(self):
        building_dataset_generator = load_torch_dataset('smart')
        for building_name, building in building_dataset_generator:
            print(f'building {building_name} dataset length: {len(building)}')
            # create a dataloader for the building
            building_dataloader = torch.utils.data.DataLoader(building, batch_size=256, shuffle=False)
            for sample in building_dataloader:
                x = sample['load']
                break
            break

    
    def test_load_pandas_datasets_direct(self):
        dataset_path = Path(os.environ['BUILDINGS_BENCH'])
        
        building_files = ['BDG-2/Bear_clean=2016', 'BDG-2/Bear_clean=2017']

        datasets = PandasBuildingDatasetsFromCSV(
                    dataset_path,
                    building_files,
                    [23.123, -32.1234],
                    BuildingTypes.COMMERCIAL,
                    features='transformer')

        for building_name, building_dset in datasets:
            print(f'building {building_name}, shape {len(building_dset)}')
            break

        datasets = PandasBuildingDatasetsFromCSV(
                    dataset_path,
                    building_files,
                    [23.123, -32.1234],
                    BuildingTypes.COMMERCIAL,
                    features='engineered'
                    )

        for building_name, building_dset in datasets:
            print(f'building {building_name}, shape {len(building_dset)}')
            break

    def test_load_pandas_datasets(self):
        
        datasets = load_pandas_dataset('bdg-2:bear')

        for building_name, building_dset in datasets:
            print(f'building {building_name}, shape {len(building_dset)}')
            break


if __name__ == '__main__':
    unittest.main()



import abc
from typing import Tuple, Dict, Union
from pathlib import Path
import torch
import torch.nn as nn


class BaseModel(nn.Module, metaclass=abc.ABCMeta):
    """Base class for all models."""
    def __init__(self, context_len, pred_len, continuous_loads):
        """Init method for BaseModel.

        Args:
            context_len (int): length of context window
            pred_len (int): length of prediction window
            continuous_loads (bool): whether to use continuous load values
        """
        super().__init__()
        self.context_len = context_len
        self.pred_len = pred_len
        self.continuous_loads = continuous_loads

    @abc.abstractmethod
    def forward(self, x: Dict) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass. 
        
        Expected keys in x:

            - 'load': torch.Tensor of shape (batch_size, seq_len, 1)
            - 'building_type': torch.LongTensor of shape (batch_size, seq_len, 1)
            - 'day_of_year': torch.FloatTensor of shape (batch_size, seq_len, 1)
            - 'hour_of_day': torch.FloatTensor of shape (batch_size, seq_len, 1)
            - 'day_of_week': torch.FloatTensor of shape (batch_size, seq_len, 1)
            - 'latitude': torch.FloatTensor of shape (batch_size, seq_len, 1)
            - 'longitude': torch.FloatTensor of shape (batch_size, seq_len, 1)

        Args:
            x (Dict): dictionary of input tensors
        Returns:
            predictions, distribution parameters (Tuple[torch.Tensor, torch.Tensor]): outputs
        """
        raise NotImplementedError()
    

    @abc.abstractmethod
    def loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """A function for computing the loss.
        
        Args:
            x (torch.Tensor): preds of shape (batch_size, seq_len, 1)
            y (torch.Tensor): targets of shape (batch_size, seq_len, 1)
        Returns:
            loss (torch.Tensor): scalar loss
        """
        raise NotImplementedError()

    @abc.abstractmethod 
    def predict(self, x: Dict) -> Tuple[torch.Tensor, torch.Tensor]:
        """A function for making a forecast on x with the model.

        Args:
            x (Dict): dictionary of input tensors
        Returns:
            predictions (torch.Tensor): of shape (batch_size, pred_len, 1)
            distribution_parameters (torch.Tensor): of shape (batch_size, pred_len, -1)
        """
        raise NotImplementedError()


    @abc.abstractmethod
    def unfreeze_and_get_parameters_for_finetuning(self):
        """For transfer learning. 
        
        - Set requires_grad=True for parameters being fine-tuned (if necessary)
        - Return the parameters that should be fine-tuned.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def load_from_checkpoint(self, checkpoint_path: Union[str, Path]):
        """Describes how to load the model from checkpoint_path."""
        raise NotImplementedError()



import unittest
from buildings_bench import transforms
from pathlib import Path
import os
import numpy as np
import torch
import pandas as pd


class TestStandardScaler(unittest.TestCase):
    def setUp(self):

        self.ss = transforms.StandardScalerTransform()
        save_dir = os.environ.get('BUILDINGS_BENCH', '')
        self.ss.load(Path(save_dir) / 'metadata' / 'transforms')


    def test_load_standard_scaler(self):

    
        self.assertIsNotNone(self.ss.mean_, True)
        self.assertIsNotNone(self.ss.std_, True)


    def test_standard_scale(self):
        
        x = torch.FloatTensor([[100.234], [0.234], [55.523]])
    
        y = self.ss.transform(x)
        z = self.ss.undo_transform(y)

        self.assertTrue(torch.allclose(x, z, atol=1e-3))


class TestBoxCox(unittest.TestCase):
    def setUp(self):

        self.bc = transforms.BoxCoxTransform()
        metadata_dir = os.environ.get('BUILDINGS_BENCH', '')
        self.bc.load(Path(metadata_dir) / 'metadata' / 'transforms')


    def test_load_boxcox(self):
        self.assertIsNotNone(self.bc.boxcox.lambdas_, True)


    def test_boxcox(self):
            
        x = torch.FloatTensor([[100.234], [0.234], [55.523]])
    
        y = self.bc.transform(x)
        z = self.bc.undo_transform(torch.from_numpy(y).float())

        # assert allclose
        self.assertTrue(torch.allclose(x, z, atol=1e-3))


class TestLatLonTransform(unittest.TestCase):
    def setUp(self):
        self.ll = transforms.LatLonTransform()

    def test_load_latlon(self):
        self.assertIsNotNone(self.ll.lat_means, True)
        self.assertIsNotNone(self.ll.lon_means, True)
        self.assertIsNotNone(self.ll.lat_stds, True)
        self.assertIsNotNone(self.ll.lon_stds, True)

    def test_latlon(self):
        x = np.array([[100.234, 0.234], [0.234, 55.523], [55.523, 100.234]])
        y = self.ll.transform_latlon(x)
        z = self.ll.undo_transform(y)

        self.assertTrue(np.allclose(x, z, atol=1e-3))


class TestTimestampTransform(unittest.TestCase):
    def setUp(self):
        self.tt = transforms.TimestampTransform()

    def test_timestamp(self):
        x = np.array(['2016-01-01 00:00:00', '2016-01-01 01:00:00'])
        # convert x to dataframe
        x = pd.DataFrame(x, columns=['timestamp'])
        y = self.tt.transform(x.timestamp)
        z = self.tt.undo_transform(y)

        #print(x,y,z)
        self.assertEqual(z[0,0], pd.to_datetime(x.timestamp).dt.dayofyear.values[0])
        self.assertEqual(z[0,1], pd.to_datetime(x.timestamp).dt.dayofweek.values[0])
        self.assertEqual(z[0,2], pd.to_datetime(x.timestamp).dt.hour.values[0])





import re
import io
import os
from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read()

with open('README.md') as f:
    readme = f.read()

# Read the version from the __init__.py file without importing it
def read(*names, **kwargs):
    with io.open(
            os.path.join(os.path.dirname(__file__), *names),
            encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()
    
def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")
    
setup(
      name='buildings_bench',
      version=find_version('buildings_bench', '__init__.py'),
      description='Large-scale pretraining and benchmarking for short-term load forecasting.',
      author='Patrick Emami',
      author_email='Patrick.Emami@nrel.gov',
      url="https://nrel.github.io/BuildingsBench/",
      long_description=readme,
      long_description_content_type='text/markdown',
      install_requires=requirements,
      packages=find_packages(include=['buildings_bench',
                                      'buildings_bench.data',
                                      'buildings_bench.evaluation',
                                      'buildings_bench.models'],
                             exclude=['test']),
      package_data={'buildings_bench': ['configs/*.toml']},
      license='BSD 3-Clause',
      python_requires='>=3.8',
      extras_require={
            'benchmark': ['transformers', 'wandb', 'properscoring', 'matplotlib', 'seaborn', 'jupyterlab']
      },
      keywords=['forecasting', 'energy', 'buildings', 'benchmark'],
      classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: BSD License",
            "Natural Language :: English",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
        ]
)



from typing import Callable
import torch 
import abc

class MetricType:
    """Enum class for metric types.

    Attributes:
        SCALAR (str): A scalar metric.
        HOUR_OF_DAY (str): A metric that is calculated for each hour of the day.
    """
    SCALAR = 'scalar'
    HOUR_OF_DAY = 'hour_of_day'


class BuildingsBenchMetric(metaclass=abc.ABCMeta):
    """An abstract class for all metrics.

    The basic idea is to acculumate the errors etc. in a list and then
    calculate the mean of the errors etc. at the end of the evaluation.

    Calling the metric will add the error to the list of errors. Calling `.mean()`
    will calculate the mean of the errors, populating the `.value` attribute.


    Attributes:
        name (str): The name of the metric.
        type (MetricType): The type of the metric.
        value (float): The value of the metric.
    """
    def __init__(self, name: str, type: MetricType):
        self.name = name
        self.type = type
        self.value = None

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        raise NotImplementedError()
    
    @abc.abstractmethod
    def reset(self):
        raise NotImplementedError()
    
    @abc.abstractmethod
    def mean(self):
        raise NotImplementedError()
 

class Metric(BuildingsBenchMetric):
    """A class that represents an error metric.  

    Example:
    
    ```python
    rmse = Metric('rmse', MetricType.SCALAR, squared_error, sqrt=True)
    mae = Metric('mae', MetricType.SCALAR, absolute_error)
    nmae = Metric('nmae', MetricType.SCALAR, absolute_error, normalize=True)
    cvrmse = Metric('cvrmse', MetricType.SCALAR, squared_error, normalize=True, sqrt=True)
    nmbe = Metric('nmbe', MetricType.SCALAR, bias_error, normalize=True)
    ```
    """
    def __init__(self, name: str, type: MetricType, function: Callable, **kwargs):
        """
        Args:
            name (str): The name of the metric.
            type (MetricType): The type of the metric.
            function (Callable): A function that takes two tensors and returns a tensor.
        
        Keyword Args:
            normalize (bool): Whether to normalize the error.
            sqrt (bool): Whether to take the square root of the error.
        """
        super().__init__(name, type)
        self.function = function
        self.kwargs = kwargs
        self.global_values = []
        self.errors = []
        self.UNUSED_FLAG = True

    def __call__(self, y_true, y_pred) -> None:
        """
        Args:
            y_true (torch.Tensor): shape [batch_size, pred_len]
            y_pred (torch.Tensor): shape [batch_size, pred_len]
        """
        self.UNUSED_FLAG = False
        self.errors += [self.function(y_true, y_pred)]
        self.global_values += [y_true]

    def reset(self) -> None:
        """Reset the metric."""
        self.global_values = []
        self.errors = []
        self.value = None
        self.UNUSED_FLAG = True

    def mean(self) -> None:
        """Calculate the mean of the error metric."""
        if self.UNUSED_FLAG:
            # Returning a number >= 0 is undefined,
            # because this metric is unused. -1
            # is a flag to indicate this.
            return
        
        # When we concatenate errors and global values
        # we want shape errors to be shape [batches, pred_len]
        # and global values to be 1D
        if self.errors[0].dim() == 1:
            self.errors = [e.unsqueeze(0) for e in self.errors]
        if self.global_values[0].dim() == 0:
            self.global_values = [g.unsqueeze(0) for g in self.global_values]
    
        all_errors = torch.concatenate(self.errors,0)
        if self.type == MetricType.SCALAR:
            mean = torch.mean(all_errors)
        elif self.type == MetricType.HOUR_OF_DAY:
            mean = torch.mean(all_errors, dim=0)
        # for root mean error
        if self.kwargs.get('sqrt', False):
            mean = torch.sqrt(mean)
        # normalize
        if self.kwargs.get('normalize', False):
            mean = mean / torch.mean(torch.concatenate(self.global_values,0))
        self.value = mean
    
    
################## METRICS ##################

def absolute_error(y_true: torch.Tensor, y_pred: torch.Tensor) -> torch.Tensor:
    """A PyTorch method that calculates the absolute error (AE) metric.

    Args:
        y_true (torch.Tensor): [batch, pred_len]
        y_pred (torch.Tensor): [batch, pred_len]
    
    Returns:
        error (torch.Tensor): [batch, pred_len]
    """
    return torch.abs(y_true - y_pred)


def squared_error(y_true: torch.Tensor, y_pred: torch.Tensor) -> torch.Tensor:
    """A PyTorch method that calculates the squared error (SE) metric.

    Args:
        y_true (torch.Tensor): [batch, pred_len]
        y_pred (torch.Tensor): [batch, pred_len]
    
    Returns:
        error (torch.Tensor): [batch, pred_len]
    """
    return torch.square(y_true - y_pred)

 
def bias_error(y_true: torch.Tensor, y_pred: torch.Tensor) -> torch.Tensor:
    """A PyTorch method that calculates the bias error (BE) metric.

    Args:
        y_true (torch.Tensor): [batch, pred_len]
        y_pred (torch.Tensor): [batch, pred_len]
    
    Returns:
        error (torch.Tensor): [batch, pred_len]    
    """
    return y_true - y_pred


from pathlib import Path
import torch
import pandas as pd
import numpy as np
from typing import List, Union, Iterator, Tuple
import buildings_bench.transforms as transforms
from buildings_bench.transforms import BoxCoxTransform, StandardScalerTransform
from buildings_bench import BuildingTypes
import pyarrow.parquet as pq


class TorchBuildingDataset(torch.utils.data.Dataset):
    """PyTorch Dataset for a single building's Pandas Dataframe with a timestamp index and a 'power' column.

    Used to iterate over mini-batches of 192-hour subsequences.
    """
    def __init__(self, 
                dataframe: pd.DataFrame,
                building_latlon: List[float],
                building_type: BuildingTypes,
                context_len: int = 168,
                pred_len: int = 24,
                sliding_window: int = 24,
                apply_scaler_transform: str = '',
                scaler_transform_path: Path = None,
                is_leap_year = False):
        """
        Args:
            dataframe (pd.DataFrame): Pandas DataFrame with a timestamp index and a 'power' column.
            building_latlon (List[float]): Latitude and longitude of the building.
            building_type (BuildingTypes): Building type for the dataset.
            context_len (int, optional): Length of context. Defaults to 168.
            pred_len (int, optional): Length of prediction. Defaults to 24.
            sliding_window (int, optional): Stride for sliding window to split timeseries into test samples. Defaults to 24.
            apply_scaler_transform (str, optional): Apply scaler transform {boxcox,standard} to the load. Defaults to ''.
            scaler_transform_path (Path, optional): Path to the pickled data for BoxCox transform. Defaults to None.
            is_leap_year (bool, optional): Is the year a leap year? Defaults to False.
        """
        self.df = dataframe        
        self.building_type = building_type
        self.context_len = context_len
        self.pred_len = pred_len
        self.sliding_window = sliding_window
        self.apply_scaler_transform = apply_scaler_transform

        self.normalized_latlon = transforms.LatLonTransform().transform_latlon(building_latlon)
        self.time_transform = transforms.TimestampTransform(is_leap_year=is_leap_year)
        if self.apply_scaler_transform == 'boxcox':
            self.load_transform = BoxCoxTransform()
            self.load_transform.load(scaler_transform_path)
        elif self.apply_scaler_transform == 'standard':
            self.load_transform = StandardScalerTransform()
            self.load_transform.load(scaler_transform_path)
        
    def __len__(self):
        return (len(self.df) - self.context_len - self.pred_len) // self.sliding_window

    def __getitem__(self, idx):
        seq_ptr = self.context_len + self.sliding_window * idx

        load_features = self.df['power'].iloc[seq_ptr-self.context_len : seq_ptr+self.pred_len].values.astype(np.float32)
        if self.apply_scaler_transform != '':
            load_features = self.load_transform.transform(load_features)
        time_features = self.time_transform.transform(self.df.index[seq_ptr-self.context_len : seq_ptr+self.pred_len ])
        latlon_features = self.normalized_latlon.reshape(1,2).repeat(self.context_len + self.pred_len, axis=0).astype(np.float32) 
        if self.building_type == BuildingTypes.RESIDENTIAL:
            building_features = BuildingTypes.RESIDENTIAL_INT * np.ones((self.context_len + self.pred_len,1), dtype=np.int32)
        elif self.building_type == BuildingTypes.COMMERCIAL:
            building_features = BuildingTypes.COMMERCIAL_INT * np.ones((self.context_len + self.pred_len,1), dtype=np.int32)

        sample = {
            'latitude': latlon_features[:, 0][...,None],
            'longitude': latlon_features[:, 1][...,None],
            'day_of_year': time_features[:, 0][...,None],
            'day_of_week': time_features[:, 1][...,None],
            'hour_of_day': time_features[:, 2][...,None],
            'building_type': building_features,
            'load': load_features[...,None]
        }
        return sample
    

class TorchBuildingDatasetFromParquet:
    """Generate PyTorch Datasets out of Parquet files.
    
    Each file has multiple buildings (with same Lat/Lon and building type) and
    each building is a column. All time series are for the same year.     

    Attributes:
        building_datasets (dict): Maps unique building ids to a TorchBuildingDataset.   
    """
    def __init__(self,
                parquet_datasets: List[str],
                building_latlons: List[List[float]],
                building_types: List[BuildingTypes],
                context_len: int = 168,
                pred_len: int = 24,
                sliding_window: int = 24,
                apply_scaler_transform: str = '',
                scaler_transform_path: Path = None,
                leap_years: List[int] = None):
        """
        Args:
            parquet_datasets (List[str]): List of paths to a parquet file, each has a timestamp index and multiple columns, one per building.
            building_latlons (List[List[float]]): List of latlons for each parquet file.
            building_types (List[BuildingTypes]): List of building types for each parquet file.
            context_len (int, optional): Length of context. Defaults to 168.
            pred_len (int, optional): Length of prediction. Defaults to 24.
            sliding_window (int, optional): Stride for sliding window to split timeseries into test samples. Defaults to 24.
            apply_scaler_transform (str, optional): Apply scaler transform {boxcox,standard} to the load. Defaults to ''.
            scaler_transform_path (Path, optional): Path to the pickled data for BoxCox transform. Defaults to None.
            leap_years (List[int], optional): List of leap years. Defaults to None.
        """
        self.building_datasets = {}

        for parquet_data, building_latlon, building_type in zip(parquet_datasets, building_latlons, building_types):
            df = pq.read_table(parquet_data)

            # Order by timestamp
            df = df.to_pandas().sort_values(by='timestamp')
            # Set timestamp as the index
            df.set_index('timestamp', inplace=True)
            df.index = pd.to_datetime(df.index, format='%Y-%m-%d %H:%M:%S')

            # split into multiple dataframes by column, keeping the index
            dfs = np.split(df, df.shape[1], axis=1)

            # for each column in the multi_building_dataset, create a BuildingYearDatasetFromCSV
            for building_dataframe in dfs:
                building_name = building_dataframe.columns[0]
                year = building_dataframe.index[0].year
                is_leap_year = True if year in leap_years else False

                # remove the year for aggregating over all years, later
                b_file = f'{building_type}_{Path(parquet_data).stem}/{building_name}'
                #by_file = f'{building_type}_{Path(parquet_data).stem}/{building_name}_year={year}'

                # rename column 1 to power
                building_dataframe.rename(columns={building_dataframe.columns[0]: 'power'}, inplace=True)
                self.building_datasets[b_file] = TorchBuildingDataset(building_dataframe, 
                                                                    building_latlon, 
                                                                    building_type, 
                                                                    context_len, 
                                                                    pred_len, 
                                                                    sliding_window,
                                                                    apply_scaler_transform,
                                                                    scaler_transform_path,
                                                                    is_leap_year)

        
    def __iter__(self) -> Iterator[Tuple[str, TorchBuildingDataset]]:
        """Generator to iterate over the building datasets.

        Yields:
            A pair of building id, TorchBuildingDataset objects. 
        """
        for building_name, building_dataset in self.building_datasets.items():
            yield (building_name, building_dataset)


class TorchBuildingDatasetsFromCSV:
    """TorchBuildingDatasetsFromCSV
    
    Generate PyTorch Datasets from a list of CSV files.

    Attributes:
        building_datasets (dict): Maps unique building ids to a list of tuples (year, TorchBuildingDataset). 
    """
    def __init__(self,
                data_path: Path,
                building_year_files: List[str],
                building_latlon: List[float],
                building_type: BuildingTypes,
                context_len: int = 168,
                pred_len: int = 24,
                sliding_window: int = 24,
                apply_scaler_transform: str = '',
                scaler_transform_path: Path = None,
                leap_years: List[int] = None):
        """
        Args:
            data_path (Path): Path to the dataset
            building_year_files (List[str]): List of paths to a csv file, each has a timestamp index and multiple columns, one per building.
            building_type (BuildingTypes): Building type for the dataset.
            context_len (int, optional): Length of context. Defaults to 168.
            pred_len (int, optional): Length of prediction sequence for the forecasting model. Defaults to 24.
            sliding_window (int, optional): Stride for sliding window to split timeseries into test samples. Defaults to 24.
            apply_scaler_transform (str, optional): Apply scaler transform {boxcox,standard} to the load. Defaults to ''.
            scaler_transform_path (Path, optional): Path to the pickled data for BoxCox transform. Defaults to None.
            leap_years (List[int], optional): List of leap years. Defaults to None.
        """
        self.building_datasets = {}
        self.building_type = building_type

        for building_year_file in building_year_files:
            name = building_year_file.split('_')[0].split('/')[1]
            year = int(building_year_file.split('=')[1])
            is_leap_year = True if year in leap_years else False
        
            df = pd.read_csv(data_path / (building_year_file + '.csv'),
                                                index_col=0, header=0, parse_dates=True)
            df.index = pd.to_datetime(df.index, format='%Y-%m-%d %H:%M:%S')
            df = df.sort_index()
            if len(df.columns) > 1:
                bldg_names = df.columns
                # split into multiple dataframes by column, keeping the index
                dfs = np.split(df, df.shape[1], axis=1)
            else:
                bldg_names = [name]
                dfs = [df]

            # for each bldg, create a TorchBuildingDatasetFromCSV
            for bldg_name, bldg_df in zip(bldg_names, dfs):
                bldg_df.rename(columns={bldg_df.columns[0]: 'power'}, inplace=True)

                if not bldg_name in self.building_datasets:
                    self.building_datasets[bldg_name] = []

                self.building_datasets[bldg_name] += [(year, TorchBuildingDataset(bldg_df, 
                                                                                building_latlon, 
                                                                                building_type, 
                                                                                context_len, 
                                                                                pred_len, 
                                                                                sliding_window,
                                                                                apply_scaler_transform,
                                                                                scaler_transform_path,
                                                                                is_leap_year))]
                        
        
    def __iter__(self) -> Iterator[Tuple[str, torch.utils.data.ConcatDataset]]:
        """A Generator for TorchBuildingDataset objects.

        Yields:
            A tuple of the building id and a ConcatDataset of the TorchBuildingDataset objects for all years.    
        """
        for building_name, building_year_datasets in self.building_datasets.items():
            building_year_datasets = sorted(building_year_datasets, key=lambda x: x[0])
            building_dataset = torch.utils.data.ConcatDataset([
                byd[1] for byd in building_year_datasets])
            yield (building_name, building_dataset)


class PandasBuildingDatasetsFromCSV:
    """Generate Pandas Dataframes from a list of CSV files.
    
    Create a dictionary of building datasets from a list of csv files.
    Used as a generator to iterate over Pandas Dataframes for each building.
    The Pandas Dataframe contain all of the years of data for the building.

    Attributes:
        building_datasets (dict): Maps unique building ids to a list of tuples (year, Dataframe).    
    """
    def __init__(self, 
                data_path: Path,
                building_year_files: List[str],
                building_latlon: List[float],
                building_type: BuildingTypes,
                features: str = 'transformer',
                apply_scaler_transform: str = '',
                scaler_transform_path: Path = None,
                leap_years: List[int] = []):
        """        
        Args:
            data_path (Path): Path to the dataset
            building_year_files (List[str]): List of paths to a csv file, each has a timestamp index and multiple columns, one per building.
            building_type (BuildingTypes): Building type for the dataset.
            features (str, optional): Type of features to use. Defaults to 'transformer'. {'transformer','engineered'}
                'transformer' features: load, latitude, longitude, hour of day, day of week, day of year, building type
                'engineered' features are an expansive list of mainly calendar-based features, useful for traditional ML models.
            apply_scaler_transform (str, optional): Apply scaler transform {boxcox,standard} to the load. Defaults to ''.
            scaler_transform_path (Path, optional): Path to the pickled data for BoxCox transform. Defaults to None.
            leap_years (List[int], optional): List of leap years. Defaults to None.
        """
        self.building_type = building_type
        self.features = features 
        self.apply_scaler_transform = apply_scaler_transform
        self.leap_years = leap_years
        
        if self.features == 'transformer':
            self.normalized_latlon = transforms.LatLonTransform().transform_latlon(building_latlon)
            
            if self.apply_scaler_transform == 'boxcox':
                self.load_transform = BoxCoxTransform()
                self.load_transform.load(scaler_transform_path)
            elif self.apply_scaler_transform == 'standard':
                self.load_transform = StandardScalerTransform()
                self.load_transform.load(scaler_transform_path)

        self.building_datasets = {}
        
        for building_year_file in building_year_files:
            #fullname = building_year_file.split('_')[0]
            name = building_year_file.split('_')[0].split('/')[1]
            year = int(building_year_file.split('=')[1])

            # load the csv file
            df = pd.read_csv(data_path / (building_year_file + '.csv'),
                             index_col=0, header=0, parse_dates=True)
            df.index = pd.to_datetime(df.index, format='%Y-%m-%d %H:%M:%S')
            df = df.asfreq('H')
            df = df.sort_index()
            
            bldg_dfs =[]
            # is multi-building file? 
            if len(df.columns) > 1:
                bldg_names = df.columns
                # split into multiple dataframes by column, keeping the index
                bldg_dfs = np.split(df, df.shape[1], axis=1)
            else:
                bldg_names = [name]
                bldg_dfs = [df]
            
            for bldg_name,df in zip(bldg_names, bldg_dfs):
                if self.features == 'engineered':
                    self._prepare_data_with_engineered_features(bldg_name, df, year)
                elif self.features == 'transformer':
                    self._prepare_data_transformer(bldg_name, df, year)
            
    def _prepare_data_with_engineered_features(self, bldg_name, df, year):
        # rename column 1 to power
        df.rename(columns={df.columns[0]: 'power'}, inplace=True)

        # Create hour_of_day,.., etc columns
        df["hour_x"] = np.sin(np.radians((360/24) * df.index.hour))
        df["hour_y"] = np.cos(np.radians((360/24) * df.index.hour))
        df["month_x"] = np.sin(np.radians((360/12) * df.index.month))
        df["month_y"] = np.cos(np.radians((360/12) * df.index.month))
    

        # add calendar-based variables as categorical data
        # see https://colab.research.google.com/drive/1ZWpJY03xLIsUrlOzgTNHemKyLatMgKrp?usp=sharing#scrollTo=NJABd7ow5EHC
        df["day_of_week"] = df.index.weekday
        df["hour_of_day"] = df.index.hour
        df["month_of_year"] = df.index.month
        df["weekend"] = df.index.weekday.isin([5,6])
        df= pd.get_dummies(df, columns=["day_of_week", "hour_of_day", "month_of_year", "weekend"], dtype=np.int32)

        if bldg_name in self.building_datasets:
            self.building_datasets[bldg_name] += [(year,df)]
        else:
            self.building_datasets[bldg_name] = [(year,df)]


    def _prepare_data_transformer(self, bldg_name, df, year):
        is_leap_year = True if year in self.leap_years else False            
        time_transform = transforms.TimestampTransform(is_leap_year)
        
        df.rename(columns={df.columns[0]: 'load'}, inplace=True)
        if self.apply_scaler_transform != '':
            df['load'] = self.load_transform.transform(df['load'].values)

        # create a column called "latitude" and "longitude" with the normalized lat/lon
        # of the same shape as 'load'
        df["latitude"] = self.normalized_latlon[0] * np.ones(df.shape[0])
        df["longitude"] = self.normalized_latlon[1] * np.ones(df.shape[0])
        

        if self.building_type == BuildingTypes.RESIDENTIAL:
            df["building_type"] = BuildingTypes.RESIDENTIAL_INT * np.ones(df.shape[0])
        elif self.building_type == BuildingTypes.COMMERCIAL:
            df["building_type"] = BuildingTypes.COMMERCIAL_INT * np.ones(df.shape[0])


        time_features = time_transform.transform(df.index)
        df["day_of_week"] = time_features[:,1] * np.ones(df.shape[0])
        df["hour_of_day"] = time_features[:,2] * np.ones(df.shape[0])
        df["day_of_year"] = time_features[:, 0] * np.ones(df.shape[0])

        if bldg_name in self.building_datasets:
            self.building_datasets[bldg_name] += [(year,df)]
        else:
            self.building_datasets[bldg_name] = [(year,df)]


    def __iter__(self) -> Iterator[Tuple[str, pd.DataFrame]]:
        """Generator for iterating over the dataset.

        Yields:
            A pair of building id and Pandas dataframe. 
                The dataframe has all years concatenated.    
        """
        for building_id, building_dataset in self.building_datasets.items():
            building_dataset = sorted(building_dataset, key=lambda x: x[0])
            df = pd.concat([df[1] for df in building_dataset])
            # fill missing values with 0
            df = df.fillna(0) 
            yield (building_id, df)


class PandasTransformerDataset(torch.utils.data.Dataset):
    """Create a Torch Dataset out of a Pandas DataFrame.

    Used to iterate over mini-batches of 192-hour sub-sequences.
    """
    def __init__(self, 
                 df: pd.DataFrame,
                 context_len: int = 168,
                 pred_len: int = 24,
                 sliding_window: int = 24):
        """
        Args:
            df (pd.DataFrame): Pandas DataFrame with columns: load, latitude, longitude, hour of day, day of week, day of year, building type
            context_len (int, optional): Length of context.. Defaults to 168.
            pred_len (int, optional): Length of prediction sequence for the forecasting model. Defaults to 24.
            sliding_window (int, optional): Stride for sliding window to split timeseries into test samples. Defaults to 24.
        """
        self.df = df
        self.context_len = context_len
        self.pred_len = pred_len
        self.sliding_window = sliding_window

    def __len__(self):
        return (len(self.df) - self.context_len - self.pred_len) // self.sliding_window

    def __getitem__(self, idx):
        seq_ptr = self.context_len + self.sliding_window * idx
        load_features = self.df['load'].iloc[seq_ptr-self.context_len : seq_ptr+self.pred_len].values.astype(np.float32)
        building_features = self.df['building_type'].iloc[seq_ptr-self.context_len : seq_ptr+self.pred_len].values.astype(np.int32)
        latlon_features = self.df[['latitude', 'longitude']].iloc[seq_ptr-self.context_len : seq_ptr+self.pred_len].values.astype(np.float32)
        time_features = self.df[['day_of_year', 'day_of_week', 'hour_of_day']].iloc[seq_ptr-self.context_len : seq_ptr+self.pred_len].values.astype(np.float32)
        sample = {
            'latitude': latlon_features[:, 0][...,None],
            'longitude': latlon_features[:, 1][...,None],
            'day_of_year': time_features[:, 0][...,None],
            'day_of_week': time_features[:, 1][...,None],
            'hour_of_day': time_features[:, 2][...,None],
            'building_type': building_features[...,None],
            'load': load_features[...,None]
        }
        return sample


def keep_buildings(dataset_generator: Union[TorchBuildingDatasetsFromCSV, TorchBuildingDatasetFromParquet],
                     building_ids: List[str]) -> Union[TorchBuildingDatasetsFromCSV, TorchBuildingDatasetFromParquet]:
    """Remove all buildings *not* listed in building_ids from the building_datasets dictionary from the generator class.
    
    Args:
        dataset_generator (Union[TorchBuildingDatasetsFromCSV, TorchBuildingDatasetFromParquet]): Dataset generator class.
        building_ids (List[str]): List of building ids to keep.
    
    Returns:
        dataset_generator (Union[TorchBuildingDatasetsFromCSV, TorchBuildingDatasetFromParquet]): Dataset generator 
            class with only the buildings listed in building_ids.
    """
    for building_id in list(dataset_generator.building_datasets.keys()):
        if building_id not in building_ids:
            del dataset_generator.building_datasets[building_id]
    return dataset_generator
